# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build scripts using HolAdo solution, with CSV files as input.
# Usually, it is copied and adapted for specific usage.


import logging
from holado_helper.script.action_script import ActionScript
from holado_core.common.exceptions.element_exception import ElementNotFoundException
from holado_helper.script.script import Script

logger = logging.getLogger(__name__)

class AnyActionScript(ActionScript):
    def __init__(self, work_dir_path=None, actor=None):
        super().__init__(work_dir_path=work_dir_path, actor=actor)
        
        self.__action_script_by_type = {}
    
    @Script.user_working_dir.setter  # @UndefinedVariable
    def user_working_dir(self, wd):
        Script.user_working_dir.fset(self, wd)  # @UndefinedVariable
        for action_type in self.__action_script_by_type.values():
            action_type.user_working_dir = wd
    
    def get_description(self, script_command, short_description, action):
        return """{short_description}
        
        Usage:
            1. Generate a template file for the desired action with parameter '--output-template' as in example 1.
                It is possible to specify the name of the output file with parameter '-o'.
            2. Modify the generated file to define actions parameters. 
            3. Process actions described in file with parameter '-i'.
            4. See result in output file. If parameter '-o' is not specified, the output filename is generated from parameter '-i'.
        
        In template files, a default value is set for each parameter. 
        For parameters with enum values, the default value is a combination of each possible enum values separated by a '|', but only one value is possible.
        
        
        In input and output files, a parameter 'STATUS' describes the status of execution of each action:
            - 'TODO' (or ''): the action is to process (in output file, it means that execution was stopped on an error on a previous action)
            - 'SUCCESS': the action was in success
            - 'ERROR': the action was in error
            - others: the action is not processed, you can specify your own status to skip this action
        In template files, this parameter is automatically initialized with status 'TODO'.
        Any output file can be used as input file, the actions that are in status 'ERROR' or 'TODO' will be processed again.
        
        Example 1: Output template for action '{action}' (case with CSV input file)
                
                {script_command} {action} --output-template -o {action}.csv
                
        Example 2: Process action '{action}' (case with CSV input file)
                
                {script_command} {action} -i {action}.csv
        """.format(short_description=short_description, script_command=script_command, action=action)
    
    def register_action_type(self, action_type_name, action_type):
        self.__action_script_by_type[action_type_name] = action_type
    
    def _get_action_type_for_action(self, action):
        for action_type in self.__action_script_by_type.values():
            if action_type.is_matching_action(action):
                return action_type
        raise ElementNotFoundException(f"Action '{action}' is not matching any registered action type (registered action types: {list(self.__action_script_by_type.keys())})")
    
    def _get_action_type_for_action_and_input_file(self, action, file_path):
        for action_type in self.__action_script_by_type.values():
            if action_type.with_input and action_type.is_matching_action_and_input_file_path(action, file_path):
                return action_type
        raise ElementNotFoundException(f"Action '{action}' with input file '{file_path}' is not matching any registered action type (registered action types: {list(self.__action_script_by_type.keys())})")
    
    def _get_action_type_for_action_and_output_file(self, action, file_path):
        for action_type in self.__action_script_by_type.values():
            if action_type.is_matching_action_and_output_file_path(action, file_path):
                return action_type
        raise ElementNotFoundException(f"Action '{action}' with output file '{file_path}' is not matching any registered action type (registered action types: {list(self.__action_script_by_type.keys())})")
    
    def _get_action_type_for_input_file(self, file_path):
        for action_type in self.__action_script_by_type.values():
            if action_type.is_matching_input_file_path(file_path):
                return action_type
        raise ElementNotFoundException(f"Input file '{file_path}' is not matching any registered action type (registered action types: {list(self.__action_script_by_type.keys())})")
    
    def _get_action_type_for_output_file(self, file_path):
        for action_type in self.__action_script_by_type.values():
            if action_type.is_matching_output_file_path(file_path):
                return action_type
        raise ElementNotFoundException(f"Output file '{file_path}' is not matching any registered action type (registered action types: {list(self.__action_script_by_type.keys())})")
    
    def possible_actions(self):
        res = set()
        for act_type in self.__action_script_by_type.values():
            res.update(act_type.possible_actions())
        return sorted(res)
    
    def run_script(self):
        # Note: At this step it is not possible to build real output path, thus action type is defined with action and filename given by user (if given)
        if self.with_action_parameter:
            if self.args.output_template and self.args.output_filename:
                action_type = self._get_action_type_for_action_and_output_file(self.args.action, self.args.output_filename)
            elif self.args.input_filename is not None:
                action_type = self._get_action_type_for_action_and_input_file(self.args.action, self.args.input_filename)
            else:
                action_type = self._get_action_type_for_action(self.args.action)
        elif self.args.output_template and self.args.output_filename:
            action_type = self._get_action_type_for_output_file(self.args.output_filename)
        elif self.args.input_filename is not None:
            action_type = self._get_action_type_for_input_file(self.args.input_filename)
        else:
            raise ElementNotFoundException(f"Unable to define a matching registered action type with script args {self.args} (registered action types: {list(self.__action_script_by_type.keys())})")
        
        action_type.run()






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_rest.api.rest.rest_client import RestClient
from holado.common.handlers.undefined import undefined_argument, undefined_value, default,\
    default_value
import os
from holado_core.common.tools.converters.converter import Converter
from holado_rest.api.rest.rest_manager import RestManager
from holado_core.common.handlers.wait import WaitFuncResult
from datetime import datetime
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)


class HostControllerClient(RestClient):
    
    @classmethod
    def new_client(cls, use_localhost=undefined_argument, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = None
        if 'url' not in kwargs:
            if use_localhost is undefined_argument:
                env_use = os.getenv("HOLADO_USE_LOCALHOST", False)
                use_localhost = Converter.is_boolean(env_use) and Converter.to_boolean(env_use)
            
            url = os.getenv("HOLADO_HOST_CONTROLLER_URL", undefined_value)
            if url is undefined_value:
                scheme = kwargs.get('scheme', undefined_value)
                if scheme is undefined_value:
                    scheme = os.getenv("HOLADO_HOST_CONTROLLER_SCHEME", "http")
                host = kwargs.get('host', undefined_value)
                if host is undefined_value:
                    host = "localhost" if use_localhost else os.getenv("HOLADO_HOST_CONTROLLER_HOST", "holado_host_controller")
                port = kwargs.get('port', undefined_value)
                if port is undefined_value:
                    if use_localhost:
                        port = os.getenv("HOLADO_HOST_CONTROLLER_HOSTPORT", 51231)
                    else:
                        port = os.getenv("HOLADO_HOST_CONTROLLER_PORT", 51231)
                
                if port is None:
                    url = f"{scheme}://{host}"
                else:
                    url = f"{scheme}://{host}:{port}"
            kwargs['url'] = url
        
        manager = RestManager(default_client_class=HostControllerClient)
        res = manager.new_client(**kwargs)
        
        return res

    
    def __init__(self, name, url, headers=None):
        super().__init__(name, url, headers)
    
    
    # Monitoring of host-controller
    
    def is_healthy(self):
        try:
            response = self.get(f"health")
            return response.status_code == 200
        except:
            return False
    
    def wait_is_healthy(self, timeout_sec=300, do_raise_on_timeout=True):
        logger.info(f"Waiting for host-controller ({self.url}) to be healthy...")
        wait_context = WaitFuncResult(f"wait host-controller is healthy", self.is_healthy)
        wait_context.with_timeout(timeout_sec) \
                    .with_raise_on_timeout(do_raise_on_timeout) \
                    .with_process_in_thread(False) \
                    .redo_until(True)
        try:
            wait_context.execute()
        except Exception as exc:
            logger.error(f"Error while waiting for host-controller ({self.url}) to be healthy: {str(exc)}")
            raise
        else:
            logger.info(f"host-controller ({self.url}) is healthy")
    
    
    # Common features
    
    def get_environment_variable_value(self, var_name):
        data = [var_name]
        response = self.get(f"os/env", json=data)
        return self.response_result(response, status_ok=[200])
    
    def get_directory_filenames(self, path, extension='.yml'):
        data = {'path':path, 'extension':extension}
        response = self.get(f"os/ls", json=data)
        return self.response_result(response, status_ok=[200])
    
    
    # Manage containers
    
    def get_containers_status(self, all_=False):
        if all_:
            response = self.get("docker/container?all=true")
        else:
            response = self.get("docker/container")
        return self.response_result(response, status_ok=[200,204])
    
    def get_container_info(self, name, all_=False):
        """Get container info
        @return container info if found, else None
        """
        if all_:
            response = self.get(f"docker/container/{name}?all=true")
        else:
            response = self.get(f"docker/container/{name}")
        return self.response_result(response, status_ok=[200,204], result_on_statuses={204:None, default:None})
    
    def get_container_status(self, name, all_=False):
        """Get container status
        @return container status if found, else None
        """
        if all_:
            response = self.get(f"docker/container/{name}/status?all=true")
        else:
            response = self.get(f"docker/container/{name}/status")
        return self.response_result(response, status_ok=[200,204], result_on_statuses={204:None, default:None})
    
    def get_container_health_status(self, name, all_=False):
        """Get container health status
        @return container health status if found, else None
        """
        if all_:
            response = self.get(f"docker/container/{name}/health_status?all=true")
        else:
            response = self.get(f"docker/container/{name}/health_status")
        return self.response_result(response, status_ok=[200,204], result_on_statuses={204:None, default:None})
    
    def restart_container(self, name, wait_started=True, timeout=default_value, start_if_gone=False):
        response = self.put(f"docker/container/{name}/restart?wait_started={'true' if wait_started else 'false'}{f'&timeout={timeout}' if timeout is not default_value else ''}")
        if start_if_gone and response.status_code == 410:
            return self.start_container(name)
        else:
            return self.response_result(response, status_ok=[200,204])
    
    def start_container(self, name, wait_started=True, timeout=default_value):
        response = self.put(f"docker/container/{name}/start?wait_started={'true' if wait_started else 'false'}{f'&timeout={timeout}' if timeout is not default_value else ''}")
        return self.response_result(response, status_ok=[200,204])
    
    def stop_container(self, name, wait_stopped=True, timeout=default_value, raise_if_gone=True):
        response = self.put(f"docker/container/{name}/stop?wait_stopped={'true' if wait_stopped else 'false'}{f'&timeout={timeout}' if timeout is not default_value else ''}")
        if not raise_if_gone and response.status_code == 410:
            return None
        else:
            return self.response_result(response, status_ok=[200,204])
    
    def wait_container(self, name, timeout=default_value, raise_if_gone=True):
        response = self.put(f"docker/container/{name}/wait{f'?timeout={timeout}' if timeout is not default_value else ''}")
        if not raise_if_gone and response.status_code == 410:
            return None
        else:
            return self.response_result(response, status_ok=[200,204])
    
    def await_container_is_started(self, name, timeout=default_value):
        if isinstance(timeout, int):
            response = self.put(f"docker/container/{name}/await_started?timeout={timeout}")
        else:
            response = self.put(f"docker/container/{name}/await_started")
        return self.response_result(response, status_ok=[200,204])
    
    def await_container_status(self, name, status='running', timeout=default_value):
        if isinstance(timeout, int):
            response = self.put(f"docker/container/{name}/await_status?status={status}&timeout={timeout}")
        else:
            response = self.put(f"docker/container/{name}/await_status?status={status}")
        return self.response_result(response, status_ok=[200,204])
    
    def await_container_health_status(self, name, status='healthy', timeout=default_value):
        if isinstance(timeout, int):
            response = self.put(f"docker/container/{name}/await_health_status?status={status}&timeout={timeout}")
        else:
            response = self.put(f"docker/container/{name}/await_health_status?status={status}")
        return self.response_result(response, status_ok=[200,204])
    
    
    # Manage logs
    
    def get_container_logs(self, name, with_timestamps=True, since=undefined_argument, until=undefined_argument):
        data = {}
        if with_timestamps is not None:
            data['with_timestamps'] = with_timestamps
        if since not in [undefined_argument, None]:
            if isinstance(since, datetime):
                since = DateTime.datetime_2_str(since)
            data['since'] = since
        if until not in [undefined_argument, None]:
            if isinstance(until, datetime):
                until = DateTime.datetime_2_str(until)
            data['until'] = until
        
        response = self.get(f"docker/logs/{name}", json=data)
        return self.response_result(response, status_ok=[200,204], result_on_statuses={204:[], default:None})
    
    def save_containers_logs(self, destination_path, file_format=default_value,
                  container_include_patterns=undefined_argument, container_exclude_patterns=undefined_argument,
                  field_include_patterns=undefined_argument, field_exclude_patterns=undefined_argument, field_include_others_name=default_value, 
                  with_timestamps=True, timestamps_field_name=default_value, since=undefined_argument, until=undefined_argument):
        data = {'destination_path':destination_path}
        if file_format not in [default_value, None]:
            data['file_format'] = file_format
        if container_include_patterns not in [undefined_argument, None]:
            data['container_include_patterns'] = container_include_patterns
        if container_exclude_patterns not in [undefined_argument, None]:
            data['container_exclude_patterns'] = container_exclude_patterns
        if field_include_patterns not in [undefined_argument, None]:
            data['field_include_patterns'] = field_include_patterns
        if field_exclude_patterns not in [undefined_argument, None]:
            data['field_exclude_patterns'] = field_exclude_patterns
        if field_include_others_name not in [default_value, None]:
            data['field_include_others_name'] = field_include_others_name
        if with_timestamps is not None:
            data['with_timestamps'] = with_timestamps
        if timestamps_field_name and timestamps_field_name is not default_value:
            data['timestamps_field_name'] = timestamps_field_name
        if since not in [undefined_argument, None]:
            if isinstance(since, datetime):
                since = DateTime.datetime_2_str(since)
            data['since'] = since
        if until not in [undefined_argument, None]:
            if isinstance(until, datetime):
                until = DateTime.datetime_2_str(until)
            data['until'] = until
        
        response = self.put(f"docker/logs/save", json=data)
        return self.response_result(response, status_ok=[200,204], result_on_statuses={204:None, default:None})
    
    
    # Manage configuration
    
    def get_yaml_file_content(self, file_path, field_keys=None):
        data = {'file_path': file_path}
        if field_keys:
            data['field_keys'] = field_keys
        response = self.get(f"config/yaml_file", json=data)
        return self.response_result(response, status_ok=[200,204])
    
    def update_yaml_file(self, file_path, text, with_backup=True, backup_extension='.ha_bak'):
        data = {
            'file_path': file_path,
            'yaml_string': text,
            'with_backup': with_backup,
            'backup_extension': backup_extension
            }
        response = self.patch(f"config/yaml_file", json=data)
        return self.response_result(response, status_ok=[200,204])
    
    def restore_yaml_file(self, file_path, backup_extension='.ha_bak'):
        data = {
            'action': 'restore',
            'file_path': file_path,
            'backup_extension': backup_extension
            }
        response = self.put(f"config/yaml_file", json=data)
        return self.response_result(response, status_ok=[200,204])
    
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.tools.tools import Tools
import logging
from holado_report.report.builders.report_builder import ReportBuilder
from holado_xml.xml.stream_xml_file import StreamXMLFile
from holado_system.system.filesystem.file import File
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_test.common.context.context_tools import ContextTools

logger = logging.getLogger(__name__)



class DetailedScenarioReportBuilder(ReportBuilder):
    """ Detailed scenario report
    It is dedicated to have a convenient compromise between completeness and shortness.
    
    Notes: XML version of this report is mandatory by test-server as it uses CampaignManager that needs it in its campaign import process.
           For this import process, all periods must be in uncompact format, thus this format is forced in XML format management.
    """
    def __init__(self, filepath, file_format='xml', exclude_statuses=None, exclude_categories=None):
        self.__file_format = file_format.lower()
        self.__exclude_statuses = exclude_statuses
        self.__exclude_categories = exclude_categories
        
        if self.__file_format == 'xml':
            self.__file = StreamXMLFile(filepath, mode='wt')
        elif self.__file_format == 'txt':
            self.__file = File(filepath, mode='wt')
        else:
            raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'txt', 'xml')")
        
    def build(self):
        '''
        The file is built after each scenario
        '''
        pass
        
    def after_scenario(self, scenario, scenario_report=None):
        from holado_report.report.report_manager import ReportManager
        status_info = ReportManager.ScenarioTools.get_current_scenario_status_info(scenario)
        
        # Manage excluded scenarios
        if self.__exclude_statuses and status_info.validation_status in self.__exclude_statuses:
            return
        if status_info.validation_category is not None and self.__exclude_categories:
            ind = status_info.validation_category.find(' (')
            category = status_info.validation_category[:ind] if ind > 0 else status_info.validation_category
            if category in self.__exclude_categories:
                return
        
        self.__file_add_scenario(scenario, scenario_report, status_info)
        
    def after_all(self):
        # Manage file
        self.__file.close()
        
    def __file_add_scenario(self, scenario, scenario_report, status_info):
        if self.__file_format == 'xml':
            self.__file_add_scenario_xml(scenario, scenario_report, status_info)
        else:
            self.__file_add_scenario_txt(scenario, scenario_report, status_info)
        
    def __file_add_scenario_xml(self, scenario, scenario_report, status_info):
        from holado_report.report.report_manager import ReportManager
        
        self.__open_file_if_needed()
        
        data = {
            'scenario': {
                'file': ReportManager.ScenarioTools.format_scenario_short_description(scenario),
                'feature': scenario.feature.name,
                'scenario': scenario.name,
                'scenario_period': ContextTools.format_context_period(status_info.scenario_context, use_compact_format=False),
                'report': scenario_report.report_path,
                'tags': "-t " + " -t ".join(scenario.feature.tags + scenario.tags),
                }
            }
        if status_info.validation_category:
            data['scenario']['validation_category'] = status_info.validation_category
        data['scenario']['validation_status'] = status_info.validation_status
        
        failure_data = {}
        if status_info.step_failed is not None:
            failure_data['step_number'] = status_info.step_failed_nb
            failure_data['step_line'] = status_info.step_failed.line
            if status_info.step_context and status_info.step_context.start_datetime is not None:
                failure_data['step_period'] = ContextTools.format_context_period(status_info.step_context, use_compact_format=False)

            step_descr = ReportManager.StepTools.get_step_description(status_info.step_failed)
            if "\n" in step_descr:
                failure_data['step'] = "\n" + Tools.indent_string(12, step_descr) + Tools.indent_string(8, "\n")
            else:
                failure_data['step'] = step_descr
                
            step_error_message = ReportManager.StepTools.get_step_error_message(status_info.step_failed)
            if step_error_message:
                if "\n" in step_error_message:
                    failure_data['error_message'] = "\n" + Tools.indent_string(12, step_error_message) + Tools.indent_string(8, "\n")
                else:
                    failure_data['error_message'] = step_error_message
            
            step_error = ReportManager.StepTools.get_step_error(status_info.step_failed)
            if step_error and step_error != step_error_message:
                if "\n" in step_error:
                    failure_data['error'] = "\n" + Tools.indent_string(12, step_error) + Tools.indent_string(8, "\n")
                else:
                    failure_data['error'] = step_error
                    
        if status_info.scenario_error:
            if "\n" in status_info.scenario_error:
                failure_data['scenario_error'] = "\n" + Tools.indent_string(12, status_info.scenario_error) + Tools.indent_string(8, "\n")
            else:
                failure_data['scenario_error'] = status_info.scenario_error
                
        if failure_data:
            data['scenario']['failure'] = failure_data
        elif status_info.validation_status != 'Passed':
            data['scenario']['failure'] = "No step failed, it has probably failed on a missing step implementation"
            
        self.__file.write_element_dict(data, pretty=True, indent=Tools.indent_string(4, ''))
        # Add 2 empty lines for more readability
        self.__file.internal_file.writelines(['', ''])
        
    def __file_add_scenario_txt(self, scenario, scenario_report, status_info):
        from holado_report.report.report_manager import ReportManager
        
        self.__open_file_if_needed()
        
        msg_list = [f"{scenario.filename} - l.{scenario.line}"]
        msg_list.append(f"    Feature: {scenario.feature.name}")
        msg_list.append(f"    Scenario: {scenario.name}")
        msg_list.append(f"    Report: {scenario_report.report_path}")
        msg_list.append(f"    Scenario period: {ContextTools.format_context_period(status_info.scenario_context)}")
        msg_list.append(f"    Tags: -t " + " -t ".join(scenario.feature.tags + scenario.tags))
        if status_info.validation_category:
            msg_list.append(f"    Validation category: {status_info.validation_category}")
        msg_list.append(f"    Validation status: {status_info.validation_status}")
        if status_info.step_failed is not None or status_info.scenario_error:
            msg_list.append(f"    Failure:")
            if status_info.step_failed is not None:
                msg_list.append(f"        Step number-line: {status_info.step_failed_nb} - l.{status_info.step_failed.line}")
                if status_info.step_context and status_info.step_context.start_datetime is not None:
                    msg_list.append(f"        Step period: {ContextTools.format_context_period(status_info.step_context)}")
                step_descr = ReportManager.StepTools.get_step_description(status_info.step_failed)
                if "\n" in step_descr:
                    msg_list.append(f"        Step:")
                    msg_list.append(Tools.indent_string(12, step_descr))
                else:
                    msg_list.append(f"        Step: {step_descr}")
                    
                step_error = ReportManager.StepTools.get_step_error(status_info.step_failed)
                if step_error:
                    if "\n" in step_error:
                        msg_list.append(f"        Error:")
                        msg_list.append(Tools.indent_string(12, step_error))
                    else:
                        msg_list.append(f"        Error: {step_error}")
            
            if status_info.scenario_error:
                if "\n" in status_info.scenario_error:
                    msg_list.append(f"        Scenario error:")
                    msg_list.append(Tools.indent_string(12, status_info.scenario_error))
                else:
                    msg_list.append(f"        Scenario error: {status_info.scenario_error}")
        else:
            msg_list.append(f"    Failure: No step failed, it has probably failed on a missing step implementation")
        msg_list.append(f"")
        msg_list.append(f"")
            
        self.__file.writelines(msg_list)
    
    def __open_file_if_needed(self):
        if not self.__file.is_open:
            self.__file.open()
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_ui_selenium.ui.gui.selenium.finders.selenium_finder import SeleniumFinder
from holado_core.common.finders.tools.find_context import ContainerFindContext
from holado_core.common.finders.tools.find_parameters import FindParameters
from holado_core.common.criterias.criteria import Criteria
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado.holado_config import Config
from holado_ui_selenium.ui.gui.selenium.handlers.selenium_by import SeleniumBy
from selenium.webdriver.common.by import By
from holado_core.common.exceptions.element_exception import NoSuchElementException

logger = logging.getLogger(__name__)


class BySeleniumFinder(SeleniumFinder):
    """ Finder by selenium.
    """
    
    def __init__(self, css:str=None, xpath:str=None, criteria:Criteria=None, description=None):
        super().__init__(description)
        self.__css = css
        self.__xpath = xpath
        self.__criteria = criteria
        
        if css is None and xpath is None and criteria is None:
            raise TechnicalException("At least one of css, xpath and criteria must be defined.")
        if css is not None and xpath is not None:
            raise TechnicalException("css and xpath cannot be defined together, but criteria can be defined in addition to css or xpath.")
    
    @property
    def css(self):
        return self.__css
    
    @property
    def xpath(self):
        return self.__xpath
    
    @property
    def criteria(self):
        return self.__criteria
    
    def _find_all_container(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        res = None
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            prefix_logs = f"{self._get_indent_string_level(find_parameters)}[BySeleniumFinder({self.element_description})._find_all_container]"
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{prefix_logs} -> begin in [{find_context.get_input_description()}]")
        
        # Prepare analyze time spent
        if find_parameters.analyze_time_spent:
            start = Tools.timer_s()
        
        # Find elements
        if self.css is not None:
            res = self.internal_api.find_elements_by(SeleniumBy(By.CSS_SELECTOR, self.css), self.criteria, find_context, find_parameters)
        elif self.xpath is not None:
            res = self.internal_api.find_elements_by(SeleniumBy(By.XPATH, self.xpath), self.criteria, find_context, find_parameters)
        else:
            criteria_context = find_context.get_criteria_context()
            criteria_parameters = find_parameters.get_criteria_parameters()
            
            # Look if container verify criteria
            if self.criteria.validate(None, criteria_context, criteria_parameters):
                res = [find_context.container]
            
            # Search in container
            if res is None:
                res = self.internal_api.find_elements(self.criteria, find_context, find_parameters)
        
        # Analyze time spent
        if find_parameters.analyze_time_spent:
            duration = Tools.timer_s() - start
            if duration > Config.threshold_warn_time_spent_s:
                logger.warning(f"{prefix_logs} -> end (took: {duration} s)     -> {len(res)} elements")
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"{prefix_logs} -> return {len(res)} candidates:\n{self._represent_candidates_output(res, 4)}")
        
        return self._build_result_list(res, find_context, find_parameters)
    
    def _find_all_in_parents(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        if self.css is not None:
            raise TechnicalException("css is not managed in this case.")
        elif self.xpath is not None:
            raise TechnicalException("xpath is not managed in this case.")
        else:
            res = []
            
            try:
                parent = self.internal_api.find_parent(self.criteria, None, find_context, find_parameters)
            except NoSuchElementException:
                parent = None
                
            if parent is not None:
                res.add(parent)
            
            return self._build_result_list(res, find_context, find_parameters)
    
    def is_valid_element(self, element, find_context, find_parameters):
        if self.css is not None:
            raise TechnicalException("css is not managed in this case.")
        elif self.xpath is not None:
            raise TechnicalException("xpath is not managed in this case.")
        elif self.criteria is not None:
            criteria_context = find_context.get_criteria_context()
            criteria_parameters = find_parameters.get_criteria_parameters()
            return self.criteria.validate(element, criteria_context, criteria_parameters)
        else:
            return super().is_valid_element(element, find_context, find_parameters)
    
    def __str__(self):
        res_list = [super().__str__()]
        
        # Values
        if self.css is not None:
            res_list.append(f"&css='{self.css}'")
        if self.xpath is not None:
            res_list.append(f"&xpath='{self.xpath}'")
        if self.criteria is not None:
            res_list.append(f"&criteria={{{self.criteria}}}")
        
        return "".join(res_list)
    
    
    
    
    
    


import os
import sys
import argparse
import logging

here = os.path.abspath(os.path.dirname(__file__))

holado_path = os.path.join(here, "..", "..")
sys.path.insert(0, os.path.join(holado_path, "src"))

from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_report.report.analyze.execution_historic_manager import ExecutionHistoricManager
from holado_core.common.tools.tools import Tools


def print_execution_historic(args):
    if not os.path.exists(args.report_path):
        raise FunctionalException(f"Report path '{args.report_path}' doesn't exist")
    
    ehm = ExecutionHistoricManager(args.report_path)
    
    with_feature_data = len(args.feature_data_names) > 0 if args.feature_data_names else False
    with_scenario_data = len(args.scenario_data_names) > 0 if args.scenario_data_names else False
    scenario_indent = 8 if with_feature_data else 0
    
    for feat_index, feature_eh in enumerate(ehm.report_execution_historic):
        if args.feature_number > 0 and args.feature_number != feat_index + 1:
            continue
        
        if with_feature_data:
            feature_data = ehm.extract_execution_historic_data(feature_eh, args.feature_data_names)
            logging.info(f"Feature {feat_index+1}:")
            for data in feature_data:
                logging.info(f"    {data[0]}: {data[1]}")
            if with_scenario_data:
                logging.info(f"    Scenarios:")
                
        if with_scenario_data:
            for scen_index, scenario_eh in enumerate(feature_eh["scenarios"]):
                if args.scenario_number > 0 and args.scenario_number != scen_index + 1:
                    continue
                
                scenario_data = ehm.extract_execution_historic_data(scenario_eh, args.scenario_data_names)
                logging.info(Tools.indent_string(scenario_indent, f"Scenario {feat_index+1}.{scen_index+1}:"))
                for data in scenario_data:
                    logging.info(Tools.indent_string(scenario_indent, f"    {data[0]}: {data[1]}"))
                
    logging.info(f"Finished")
    
def parse_arguments(descr=None):
    # Arguments
    if descr is None:
        descr = """Display information on scenarios executed in a campaign.
        It extracted from execution_historic.json files.
        It needs as input the folder of a campaign, and it will explore all execution_historic.json files of each scenario, in order of execution.
              
        Example 1: Display feature durations
                
                python print_execution_historic.py -r {path_to_report} -f feature.name -f feature.duration
                
        Example 1: Display scenario durations of feature n°12
                
                python print_execution_historic.py -r {path_to_report} -fn 12 -f feature.name -f feature.filename -s scenario.name -s scenario.line -s scenario.duration
                
        """
    
            
    parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-r', '--report-path', dest='report_path',
                   help='Path to the test report folder.')
    
    parser.add_argument('-rd', '--report-dirname', dest='report_dirname',
                   help='Directory name of the test report folder.')
    
    parser.add_argument('-fn', '--feature-number', dest='feature_number', type=int, default=-1,
                   help="Number of the feature to print.")
    
    parser.add_argument('-sn', '--scenario-number', dest='scenario_number', type=int, default=-1,
                   help="Number of the scenario to print.")
    
    parser.add_argument('-f', '--feature-data', dest='feature_data_names', action='append',
                   help="Fullnames of a feature data to print.")
    
    parser.add_argument('-s', '--scenario-data', dest='scenario_data_names', action='append',
                   help="Fullnames of a scenario data to print.")
    
    return parser.parse_args()

if __name__ == "__main__":
    args = parse_arguments()

    logging.basicConfig(format = '{asctime:s} | {thread:-5d} | {levelname:5s} | {module:35s} | {message:s}', style = '{', level=logging.INFO, handlers=[logging.StreamHandler()])
    
    # Execute
    try:
        print_execution_historic(args)
    except FunctionalException as exc:
        logging.error(f"Update is interrupted on error: {exc}")
        sys.exit(1)
    except:
        logging.exception("An error occured")
        sys.exit(1)

    sys.exit(0)
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException, TimeoutTechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_core.common.tools.tools import Tools
import time
from holado.holado_config import Config
from holado.common.handlers.object import Object
from holado.common.handlers.undefined import undefined_argument, undefined_value

logger = logging.getLogger(__name__)

try:
    import grpc_requests
    import grpc
    with_grpc_requests = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"GRpcClient is not available. Initialization failed on error: {exc}")
    with_grpc_requests = False

# Activation of logs in gRPC library (activated logs are thrown in stdout, it doesn't follow logging settings)
# os.environ["GRPC_VERBOSITY"] = "DEBUG"
# os.environ["GRPC_TRACE"] = "http"
# os.environ["GRPC_TRACE"] = "all"


class GRpcClient(Object):
    """
    gRPC client
    
    It is implemented internally with grpc_requests, that manages reflection clients and stub clients.
    """
    
    @classmethod
    def is_available(cls):
        return with_grpc_requests
    
    if with_grpc_requests:
        def __init__(self, name, endpoint, **kwargs):
            super().__init__(name)
            self.__endpoint = endpoint
            self.__kwargs = kwargs
            self.__client = grpc_requests.client.get_by_endpoint(endpoint, **kwargs)
            
            self.__func_grpc_services = None
            self.__func_protobuf_converter = None
            self.__func_protobuf_messages = None
            
            self.__request_log_level = Config.log_level_grpc_request
        
        def initialize(self, func_grpc_services, func_protobuf_converter, func_protobuf_messages):
            self.__func_grpc_services = func_grpc_services
            self.__func_protobuf_converter = func_protobuf_converter
            self.__func_protobuf_messages = func_protobuf_messages
            
        @property
        def __grpc_services(self):
            return self.__func_grpc_services()
    
        @property
        def __protobuf_converter(self):
            return self.__func_protobuf_converter()
    
        @property
        def __protobuf_messages(self):
            return self.__func_protobuf_messages()
        
        @property    
        def internal_client(self) -> grpc_requests.client.Client:
            return self.__client
        
        @property
        def request_log_level(self):
            return self.__request_log_level
        
        @request_log_level.setter
        def request_log_level(self, log_level):
            self.__request_log_level = log_level
        
        def request(self, service, method, request, raw_output=False, result_on_statuses=undefined_argument, **kwargs):
            """
            :param request: request data in json or proto format
            :param raw_output: if method should return a proto object or a json oject (default: False)
            :param kwargs: other arguments for underlying grpc_requests method, or further underlying grpc method (ex: 'timeout')
            :returns: if raw_output==True, returns a proto object, else returns a json object with proto data
            """
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Requesting {service}.{method} with data [{request}] (raw_output: {raw_output} ; kwargs:{kwargs})")
            res = undefined_value
            
            # Set a default timeout
            raise_on_timeout = False
            if kwargs is None:
                kwargs = {}
            if 'timeout' not in kwargs:
                kwargs['timeout'] = Config.join_timeout_seconds
                raise_on_timeout = True
            timeout = kwargs['timeout'] if 'timeout' in kwargs else None
            
            success = False
            last_exc = None
            for try_nb in range(1,4):
                try:
                    # Ask always raw_output=True, so that we get a Protobuf instance, and then conversion is done if needed
                    res_proto = self.internal_client.request(service, method, request, raw_output=True, **kwargs)
                except Exception as exc:
                    last_exc = exc
                    exc_str = str(exc)
                    
                    # Manage result on statuses
                    if result_on_statuses is not undefined_argument:
                        for status, status_result in result_on_statuses.items():
                            if f"status = StatusCode.{status.name}" in exc_str:
                                res = status_result
                                break
                        if res is not undefined_value:
                            success = True
                            break
                    
                    msg_list = [
                        f"Request failed (try {try_nb}):",
                        f"    method: {service}.{method}",
                        f"    data:",
                        Tools.indent_string(8, str(request)),
                        f"    raw_output: {raw_output}",
                        f"    kwargs: {kwargs}",
                        f"    error:",
                        Tools.indent_string(8, exc_str) ]
                    exc_msg = "\n".join(msg_list)
                    if "status = StatusCode.UNAVAILABLE" in exc_str:
                        logger.warning("Service temporarily unavailable:\n" + exc_msg)
                        time.sleep(1)
                        continue
                    elif "status = StatusCode.DEADLINE_EXCEEDED" in exc_str:
                        if raise_on_timeout:
                            raise TimeoutTechnicalException(exc_msg)
                        else:
                            logger.warning(f"Timeout ({timeout} s) occured while requesting:\n" + exc_msg)
                            time.sleep(1)
                            continue
                    elif "status = " in exc_str:
                        raise FunctionalException(exc_msg) from exc
                    else:
                        raise TechnicalException(exc_msg) from exc
                else:
                    success = True
                    break
            
            # If still in error after all tries, raise an exception with last exception message
            if not success:
                if "status = " in exc_msg:
                    raise FunctionalException(exc_msg) from last_exc
                else:
                    raise TechnicalException(exc_msg) from last_exc
                
            # Manage result conversion if needed
            # Note: this step is done manually since grpc_requests has some limitations when raw_output=False:
            #     - Field with default values are not set in json result
            #     - Some field types are badly managed (ex: uint64 fields appear as string in json)
            if res is undefined_value:
                if raw_output == True:
                    if isinstance(res_proto, grpc._channel._MultiThreadedRendezvous):
                        res = list(res_proto)
                    else:
                        res = res_proto
                else:
                    if isinstance(res_proto, grpc._channel._MultiThreadedRendezvous):
                        res = [self.__protobuf_converter.convert_protobuf_object_to_json_object(cur_res) for cur_res in res_proto]
                    else:
                        res = self.__protobuf_converter.convert_protobuf_object_to_json_object(res_proto)
            if Tools.do_log(logger, self.request_log_level):
                logger.log(self.request_log_level, f"Request {service}.{method} with data [{request}] (raw_output: {raw_output} ; kwargs:{kwargs})  => {res}")
            return res
                
        def get_request_data_type_fullname(self, service, method):
            method_descriptor = self.__grpc_services.get_method_descriptor(service, method)
            return method_descriptor.input_type.full_name
            
        def build_request_data(self, service, method, params_table=None, params_dict=None, as_proto=False):
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Building request data for service method '{service}.{method}'")
            if as_proto is not None and as_proto:
                method_descriptor = self.__grpc_services.get_method_descriptor(service, method)
                res = self.__protobuf_messages.new_message(method_descriptor.input_type.full_name, fields_table=params_table, fields_dict=params_dict)
            else:
                if params_table is not None:
                    res = ValueTableConverter.convert_name_value_table_2_json_object(params_table)
                elif params_dict is not None:
                    res = params_dict
                else:
                    res = {}
            
            return res
        
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_json.ipc.json_types import JsonTypes
from holado_core.common.handlers.features.resource_by_type import FeatureObjectAndClassResourceByType
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)

class JsonConverter(FeatureObjectAndClassResourceByType):
    
    def __init__(self):
        super().__init__()
        self.import_registered_resources_in_class(class_=JsonTypes)
    
    @classmethod
    def _get_class_resource_description(cls, plural=False):
        if plural:
            return 'class json types'
        else:
            return 'class json type'
    
    def _get_resource_description(self, plural=False):
        if plural:
            return 'json types'
        else:
            return 'json type'
    
    def to_json(self, obj):
        resource = self.get_resource_for_type(obj=obj)
        if resource[1] is not None:
            return resource[1](obj)
        else:
            raise TechnicalException(f"Resource 'to_json' is not defined for object of type {Typing.get_object_class_fullname(obj)}")
    
    def from_json_to(self, obj_json, class_, _parents=None):
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{Typing.get_object_class_fullname(self)}] from_json_to({obj_json}, {Typing.get_class_fullname(class_)})")
        # Find resource associated to class_, else find the resource associated to json object
        resource = None
        if class_ is not None:
            resource = self.get_resource_for_type(class_=class_, raise_if_not_found=False)
        if resource is None:
            resource = self.get_resource_for_type(obj=obj_json)
        
        # If defined, use complex json conversion method, else use simple method
        if resource[2] is not None:
            return resource[2](obj_json, class_, _parents=_parents, converter=self)
        elif resource[0] is not None:
            return resource[0](obj_json)
        else:
            raise TechnicalException(f"Resource 'from_json_to' is not defined for type {class_}")
        
    
    
    
    



# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import hashlib
from holado_core.common.exceptions.technical_exception import TechnicalException

logger = logging.getLogger(__name__)




class HashTools(object):
    
    @classmethod
    def code(cls, content, code_size=None, prefix=None):
        hex_size = code_size
        if hex_size is not None:
            if prefix is not None:
                hex_size = hex_size - len(prefix)
            hex_size = hex_size // 2 * 2
        
        res = cls.hexblake2(content, hex_size=hex_size)
        
        if prefix is not None:
            res = prefix + res
        
        return res
    
    @classmethod
    def hash(cls, algorithm_name, content, hash_size=None, **hash_kwargs):
        hash_obj = hashlib.new(algorithm_name, **hash_kwargs)
        cls._update_hash(hash_obj, content)
        res = hash_obj.digest()
        
        # Truncate result size if needed
        if hash_size is not None:
            if not isinstance(hash_size, int):
                raise TechnicalException(f"Parameter size must be an integer")
            res = res[:hash_size]
        
        return res
    
    @classmethod
    def hexhash(cls, algorithm_name, content, hex_size=None, **hash_kwargs):
        hash_obj = hashlib.new(algorithm_name, **hash_kwargs)
        cls._update_hash(hash_obj, content)
        res = hash_obj.hexdigest()
        
        # Truncate result size if needed
        if hex_size is not None:
            if not isinstance(hex_size, int):
                raise TechnicalException(f"Parameter hex_size must be an integer")
            res = res[:hex_size]
        
        return res
        
    @classmethod
    def hexmd5(cls, content):
        hash_obj = hashlib.md5()
        cls._update_hash(hash_obj, content)
        return hash_obj.hexdigest()
    
    @classmethod
    def hexblake2(cls, content, hex_size=None, **hash_kwargs):
        if hex_size is not None:
            if isinstance(hex_size, int):
                digest_size = hex_size // 2
            else:
                raise TechnicalException(f"Parameter hex_size must be an integer")
            hash_kwargs['digest_size'] = digest_size
        
        hash_obj = hashlib.blake2b(**hash_kwargs)
        cls._update_hash(hash_obj, content)
        return hash_obj.hexdigest()
    
    @classmethod
    def _update_hash(cls, hash_obj, content):
        # Manage list separately for memory purpose
        done = False
        if not isinstance(content, str): 
            try:
                for cont in content:
                    if type(cont) != bytes:
                        cont = cont.encode('utf-8')
                    hash_obj.update(cont)
                done = True
            except TypeError:
                done = False
    
        if not done:
            if type(content) != bytes:
                content = content.encode('utf-8')
            hash_obj.update(content)
    
    
    
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_report.report.report_manager import BaseReport
from holado_core.common.exceptions.technical_exception import TechnicalException
import weakref
# from holado_core.scenario.scenario_duration_manager import ScenarioDurationManager

logger = logging.getLogger(__name__)



class FeatureReport(BaseReport):
    TScenarioReport = None
    
    def __init__(self, report_manager, feature_context, feature):
        super().__init__()
        
        self.__report_manager_weakref = weakref.ref(report_manager)
        self.__feature_context = feature_context
        self.__feature = feature
        
        # Auto configuration
        self.configure()
        
    def configure(self):
        from holado_report.report.reports.scenario_report import ScenarioReport
        FeatureReport.TScenarioReport = ScenarioReport
        
    def initialize_reports(self):
    #     self.__execution_historic = ExecutionHistoric()
    #
    #     fn = self.get_path("execution_historic.json")
    #     self.__report_builders.append(ExecutionHistoricReportBuilder(self.__execution_historic, fn))
    #
    #     fn = self.get_path("report_summary_scenario_failed.txt")
    #     self.__report_builders.append(SummaryDetailedScenarioReportBuilder(fn))
    #
    #     fn = self.get_path("report_short_scenario_failed.txt")
    #     self.__report_builders.append(ShortDetailedScenarioReportBuilder(fn))
    #
    #     # fn = self.get_path("report_detailed_scenario_failed.txt")
    #     fn = self.get_path("report_detailed_scenario_failed.xml")
    #     self.__report_builders.append(DetailedScenarioReportBuilder(fn))
    #
    #     fn = self.get_path("report_summary.txt")
    #     self.__report_builders.append(SummaryReportBuilder(self.__execution_historic, fn))
        
        # Initialize reports
        self.before_all()
        self.before_feature(self.feature_context, self.feature, self)
    
    @property
    def feature(self):
        return self.__feature
    
    @property
    def feature_context(self):
        return self.__feature_context
    
    @property
    def report_manager(self):
        return self.__report_manager_weakref()
    
    @property
    def scenario_reports(self):
        return self.children_reports("scenario")
    
    @property
    def current_scenario_report(self):
        if self.scenario_reports:
            return self.scenario_reports[-1][1]
        else:
            return None
    
    def new_scenario_report(self, scenario_context, scenario):
        res = FeatureReport.TScenarioReport(self, scenario_context, scenario)
        self.add_child_report(res, "scenario", scenario.name, f"{scenario.filename} - l.{scenario.line}")
        res.initialize_reports()
        return res
        
    def before_scenario(self, scenario_context, scenario, scenario_report=None):
        super().before_scenario(scenario_context, scenario, scenario_report)
    
    def after_scenario(self, scenario, scenario_report=None):
        if self.current_scenario_report.scenario != scenario:
            raise TechnicalException(f"Processing after scenario '{scenario.name}' whereas current scenario is '{self.current_scenario_report.scenario.name}'")
        if self.current_scenario_report is not scenario_report:
            raise TechnicalException(f"Processing after scenario report '{scenario_report}' whereas current scenario report is '{self.current_scenario_report}'")
        self.current_scenario_report.build_reports()

        super().after_scenario(scenario, scenario_report)
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import abc
from holado_core.common.finders.tools.find_info import FindInfo


class ElementHolder(object):
    """ Information on an element.
    """
    __metaclass__ = abc.ABCMeta    
    
    def __init__(self, parent, instance, description=None):
        """
        @param parent Parent
        @param instance Element instance
        @param description Element description
        """
        self.__parent = parent
        self.__instance = instance
        self.__description = description
        self.__find_info = None

    @property
    def parent(self):
        """
        @return Parent
        """
        return self.__parent
    
    @property
    def element(self):
        """
        @return Element instance
        """
        return self.__instance

    @property
    def description(self):
        """
        @return Element description
        """
        return self.__description

    @description.setter
    def description(self, description):
        """
        @param description New description
        """
        self.__description = description

    @property
    def complete_description(self):
        """
        @return Element description with parent descriptions
        """
        if self.parent is not None:
            return f"{self.parent.get_complete_description()} -> {self.description}"
        else:
            return self.description

    @property
    def complete_description_and_details(self):
        """
        @return Element description with parent descriptions and details on element
        """
        return f"{self.complete_description} [{self.element}]"
    
    @property
    def find_info(self):
        """
        @return Find information
        """
        return self.__find_info

    def update_find_info(self, finder, find_context, find_parameters):
        """
        Update find information
        @param finder Finder used to find element
        @param find_context Find context
        @param find_parameters Find parameters
        """
        if self.__find_info is None:
            findInfo = FindInfo()
            findInfo.finder = finder
            findInfo.find_context = find_context
            findInfo.find_parameters = find_parameters
    
    def get_holder_for(self, element_instance, element_description):
        """
        @param element_instance New element instance.
        @param element_description New element description.
        @return Element holder for given element instance
        """
        raise NotImplementedError
    
    def get_list_holder_for(self, element_instances, element_description):
        """
        @param element_instances New element instance list.
        @param element_description New element description.
        @return Element holder for given element instance
        """
        res = []
        for element_instance in element_instances:
            res.append(self.get_holder_for(element_instance, element_description))
        return res
    
    def __eq__(self, obj:object)->bool:
        return self.__instance is obj.element or self.__instance == obj.element
    
    
    
    


# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from typing import NamedTuple, get_origin
from typing import types  # @UnresolvedImport
from holado_core.common.handlers.enums import AccessType
import inspect
from holado_core.common.exceptions.technical_exception import TechnicalException


logger = logging.getLogger(__name__)


class Typing(object):
    
    @classmethod
    def is_NamedTuple(cls, obj=None, class_=None):
        # if obj is not None:
        #     class_ = type(obj)
        # return hasattr(class_, '__orig_bases__') and class_.__orig_bases__[0] == NamedTuple
        if obj is not None:
            return (hasattr(obj, "__base__") and obj.__base__ == tuple or not hasattr(obj, "__dict__")) \
                and hasattr(obj, '_fields') and len(obj._fields) > 0 and any(map(lambda x:isinstance(x, str), obj._fields))
        else:
            return hasattr(class_, '__orig_bases__') and class_.__orig_bases__[0] == NamedTuple
    
    @classmethod
    def is_Generic(cls, generic_name, obj=None, class_=None):
        if obj is not None:
            class_ = type(obj)
        return hasattr(class_, '_name') and class_._name == generic_name
    
    @classmethod
    def is_List(cls, obj=None, class_=None):
        return cls.is_Generic('List', obj, class_)
    
    @classmethod
    def is_Dict(cls, obj=None, class_=None):
        return cls.is_Generic('Dict', obj, class_)
    
    @classmethod
    def is_typing_type(cls, class_):
        return get_origin(class_) is not None
    
    @classmethod
    def get_origin_type(cls, class_):
        if class_ is not None and cls.is_typing_type(class_):
            orig_type = get_origin(class_)
            if type(orig_type) == type:
                return orig_type
            else:
                return None
        else:
            return class_
    
    @classmethod
    def is_function(cls, target):
        return isinstance(target, types.FunctionType)
    
    @classmethod
    def is_method(cls, target):
        return isinstance(target, types.MethodType)
    
    @classmethod
    def get_method_function(cls, target):
        return target.__func__
    
    @classmethod
    def get_method_object_instance(cls, target):
        return target.__self__
    
    @classmethod
    def get_function_qualname(cls, target):
        return target.__qualname__
    
    @classmethod
    def get_function_fullname(cls, target):
        func_qualname = cls.get_function_qualname(target)
        if hasattr(target, "__module__"):
            return target.__module__ + "." + func_qualname
        else:
            return func_qualname
    
    
    @classmethod
    def get_class_name(cls, class_):
        return class_.__name__
    
    @classmethod
    def get_class_fullname(cls, class_):
        if class_ is None:
            return None
        
        class_name = cls.get_class_name(class_)
        if hasattr(class_, "__module__"):
            return class_.__module__ + "." + class_name
        else:
            return class_name
    
    @classmethod
    def get_object_class(cls, obj):
        if hasattr(obj, "__class__"):
            return obj.__class__
        else:
            return type(obj)
    
    @classmethod
    def get_object_class_name(cls, obj):
        return cls.get_object_class(obj).__name__
    
    @classmethod
    def get_object_class_fullname(cls, obj):
        obj_class_name = cls.get_object_class_name(obj)
        if hasattr(obj, "__module__"):
            return obj.__module__ + "." + obj_class_name
        else:
            return obj_class_name
    
    @classmethod
    def get_object_attributes(cls, obj, access_type=AccessType.Public):
        if hasattr(obj, "__module__") and getattr(obj, "__module__") == 'zeep.objects' and hasattr(obj, "__values__"):
            # Manage zeep objects separately
            d = getattr(obj, "__values__")
            return [(key, d[key]) for key in sorted(d)]
        elif Typing.is_NamedTuple(obj):
            return [(k, getattr(obj, k)) for k in obj._fields]
        else:
            try:
                try:
                    if not hasattr(obj, '__bases__'):
                        obj.__bases__ = ()
                except KeyError:
                    obj.__bases__ = ()
                except AttributeError:
                    pass
            
                if access_type == AccessType.Public:
                    return [(name, member) for name, member in inspect.getmembers(obj, lambda a: not inspect.isroutine(a)) if not name.startswith('_')]
                elif access_type == AccessType.Protected:
                    return [(name, member) for name, member in inspect.getmembers(obj, lambda a: not inspect.isroutine(a)) if not name.startswith('__')]
                elif access_type == AccessType.Private:
                    return [(name, member) for name, member in inspect.getmembers(obj, lambda a: not inspect.isroutine(a))]
                else:
                    raise TechnicalException(f"Unmanaged access type '{access_type}'")
            except ModuleNotFoundError:
                return []
    
    @classmethod
    def get_object_attribute_names(cls, obj):
        if hasattr(obj, "__module__") and getattr(obj, "__module__") == 'zeep.objects' and hasattr(obj, "__values__"):
            # Manage zeep objects separately
            d = getattr(obj, "__values__")
            return [key for key in sorted(d)]
        else:
            return [name for name, _ in inspect.getmembers(obj, lambda a: not inspect.isroutine(a)) if not name.startswith('_')]
    
    @classmethod
    def get_object_attribute_values_by_name(cls, obj):
        if hasattr(obj, "__module__") and getattr(obj, "__module__") == 'zeep.objects' and hasattr(obj, "__values__"):
            # Manage zeep objects separately
            return getattr(obj, "__values__")
        else:
            return {name: member for name, member in inspect.getmembers(obj, lambda a: not inspect.isroutine(a)) if not name.startswith('_')}
    
    @classmethod
    def get_class_for_type_name(cls, type_name):
        type_name_splitted = type_name.split('.')
        if len(type_name_splitted) > 1:
            module = __import__(type_name_splitted[0])
            for name in type_name_splitted[1:-1]:
                module = getattr(module, name)
            res = getattr(module, type_name_splitted[-1])
        else:
            res = globals()[type_name]
            
        return res

    
    





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from typing import NamedTuple, List, get_type_hints
from holado_helper.script.action import BehaveActionInfo
from holado_json.ipc.json_converter import JsonConverter
from holado_json.ipc.json_types import JsonTypes
from holado_json.ipc.json_types import _from_json_to_NamedTuple as orig_from_json_to_NamedTuple
from holado_json.ipc.json_types import _from_json_to_object as orig_from_json_to_object
from holado_python.standard_library.typing import Typing
import inspect

logger = logging.getLogger(__name__)

class Job(NamedTuple):
    actions:List[BehaveActionInfo]
    in_parallel:bool
    
    
class JobJsonConverter(JsonConverter):
    def __init__(self, script_inst):
        super().__init__()
        self.__script_inst = script_inst
        self.__job_counter = 0
        self.__action_info_counter = 0
        
        self.__override_registered_types()
        
    def __override_registered_types(self):
        self.register_resource_for_type('NamedTuple', lambda c: Typing.is_NamedTuple(class_=c), lambda o: Typing.is_NamedTuple(obj=o), 
                           (None, None, self._from_json_to_NamedTuple))

        self.register_resource_for_type_in_class('object', lambda c: inspect.isclass(c) and issubclass(c, object), lambda o: isinstance(o, object) and hasattr(o, '__dict__'), 
                                (None, lambda x: x.__dict__, self._from_json_to_object))
    
    def _from_json_to_object(self, obj_json, class_, _parents=None, converter=None):
        if inspect.isclass(class_) and issubclass(class_, BehaveActionInfo):
            self.__action_info_counter += 1
            action = obj_json['action']
            
            params = self.from_json_to(obj_json['params'], dict, _parents=JsonTypes._new_parents(_parents, class_, name='params'))
            action_params = self.__script_inst.build_processing_action_params(action, action_args=obj_json, action_params=params, job_id=self.__job_counter)
            
            static_params = self.__script_inst.action_registry.get_action_static_parameters_values(action)
            
            tags = self.__script_inst.build_action_tags(action, action_args=obj_json)
            behave_args = self.__script_inst._build_behave_args(tags)
            
            index = _parents[-1]['index'] if _parents is not None and 'index' in _parents[-1] else None
            return self.__script_inst.new_action_info(action, behave_args=behave_args, params=action_params, static_params=static_params, index=index)
        else:
            return orig_from_json_to_object(obj_json, class_, _parents=_parents, converter=converter)
    
    def _from_json_to_NamedTuple(self, obj_json, class_, _parents=None, converter=None):
        if inspect.isclass(class_) and issubclass(class_, Job):
            self.__job_counter += 1
            actions = self.from_json_to(obj_json['actions'], get_type_hints(Job)['actions'], _parents=JsonTypes._new_parents(_parents, class_, name='actions'))
            is_parallel = obj_json.get('is_parallel', False)
            return Job(actions, is_parallel)
        else:
            return orig_from_json_to_NamedTuple(obj_json, class_, _parents=_parents, converter=converter)






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.finders.tools.enums import FindType
from holado_ui.ui.gui.windows.gui_window import GUIWindow
import abc
from holado_ui_selenium.ui.gui.selenium.handlers.selenium_holder import SeleniumHolder,\
    PopupHolder
from holado_ui.ui.handlers.ui_context import UIContext
from selenium.common.exceptions import NoAlertPresentException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_ui_selenium.ui.gui.selenium.handlers.selenium_redo import SeleniumRedo
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_ui_selenium.ui.gui.selenium.tables.converters.selenium_table_converter import SeleniumTableConverter
from holado_ui_selenium.ui.gui.selenium.tables.comparators.selenium_2_value_table_comparator import Selenium2Value_TableComparator
from holado_ui_selenium.ui.gui.selenium.tables.comparators.selenium_2_value_table_with_header_comparator import Selenium2Value_TableWithHeaderComparator
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)



class SeleniumWindow(GUIWindow):
    """ Base class for Selenium windows.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self):
        super().__init__()
    
    def get_window_container(self):
        return SeleniumHolder(None, self.driver, self.driver.internal_driver, "page content")
    
    def click_on_element_by_id(self, ui_context:UIContext, element_id):
        """
        Click on element of given ID value
        @param ui_context UI Context
        @param element_id Element ID
            """
        finder = self.inspector.get_finder_by_id(element_id, self.inspect_builder().context(ui_context), None)
        self.actor.act_find_and_click_on(finder, FindType.In).execute(self.current_container)
    
    def __find_alert(self):
        """
        Checks and register an alert
                """
        res = None
        
        try:
            res = self.driver.internal_driver.swith_to.alert
        except NoAlertPresentException:
            res = None
        
        return res
    
    def check_new_popup_is_opened(self, expected_popup_title=None, raise_exception=True):
        res = True
        
        # Search if an alert is present
        alert = self.__find_alert()
        if alert is not None:
            # Checking popup title
            if expected_popup_title is not None and len(expected_popup_title) > 0:
                raise FunctionalException(f"An alert popup is present while checking popup '{expected_popup_title}'")
            
            # Update container info
            self._enter_popup("{ALERT}", SeleniumHolder(self.driver, alert, self.driver.internal_driver, "{ALERT}"))
        else:
            popup = self.find_popup(expected_popup_title, raise_exception)
            if popup is not None:
                # Update container info
                popup_name = expected_popup_title
                if popup_name is None:
                    popup_name = "{EMPTY}"
                self._enter_popup(popup_name, self._get_popup_container(expected_popup_title, popup))
            else:
                # This case appear when raise_exception is False and popup wasn't found
                res = False
        
        if not res and raise_exception:
            if expected_popup_title is not None:
                raise FunctionalException(f"Unable to find a popup of title '{expected_popup_title}'")
            else:
                raise FunctionalException("Unable to find a popup")
        
        return res
    
    def _get_popup_container(self, popup_title, popup_element):
        # Update container info
        if popup_title is not None and len(popup_title) > 0:
            popup_description = f"popup '{popup_title}'"
        else:
            popup_description = "popup"
        return PopupHolder(self.driver, self._get_web_element(popup_element), self.driver.internal_driver, popup_title, popup_description)
    
    def get_table(self, ui_context:UIContext=None):
        """
        Find a table and return it as a Table instance
        @param ui_context UI Context
        @return a Table instance
        """
        finder = self.inspector.get_finder_table(self.inspect_builder.context(ui_context), None)
        return self.get_table_by_finder(finder, FindType.In, self.current_container)
    
    def get_table_with_header(self, ui_context:UIContext=None):
        """
        Find a table and return it as a Table instance
        @param ui_context UI Context
        @return a Table instance
        """
        finder = self.inspector.get_finder_table(self.inspect_builder.context(ui_context), None)
        return self.get_table_with_header_by_finder(finder, FindType.In, self.current_container)
    
    def get_table_by_finder(self, finder_table, find_type, container):
        """
        Find a table with given finder in given container and return it as a Table instance
        @param finder_table Finder for table
        @param find_type Find type
        @param container Container
        @return Table instance
        """
        # Find table in given zone
        table_element = finder_table.find(find_type, container)
        
        # Convert to table type
        return SeleniumTableConverter.convert_element_to_selenium_table(self.inspector, table_element)
    
    def get_table_with_header_by_finder(self, finder_table, find_type, container):
        """
        Find a table with given finder in given container and return it as a Table instance
        @param finder_table Finder for table
        @param find_type Find type
        @param container Container
        @return Table instance
        """
        # Find table in given zone
        table_element = finder_table.find(find_type, container)
        
        # Convert to table type
        return SeleniumTableConverter.convert_element_to_selenium_table_with_header(self.inspector, table_element)
    
    def check_table_contains(self, ui_context, expected, raise_exception):
        class CheckRedo(SeleniumRedo):
            def __init__(self_redo):  # @NoSelf
                super().__init__("check table contains", self.driver)
            
            def _process(self_redo):  # @NoSelf
                if isinstance(expected, ValueTableWithHeader):
                    obtained = self.get_table_with_header(ui_context)
                    comparator = Selenium2Value_TableWithHeaderComparator(self.inspector)
                else:
                    obtained = self.get_table(ui_context)
                    comparator = Selenium2Value_TableComparator(inspector=self.inspector)
                return comparator.contains_rows(obtained, expected, raise_exception=raise_exception)
            
        redo = CheckRedo()
        return redo.execute()
    
    def check_table_displays(self, ui_context, expected, raise_exception):
        class CheckRedo(SeleniumRedo):
            def __init__(self_redo):  # @NoSelf
                super().__init__("check table displays", self.driver)
            
            def _process(self_redo):  # @NoSelf
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace("Comparing tables...")
                    
                if isinstance(expected, ValueTableWithHeader):
                    obtained = self.get_table_with_header(ui_context)
                    comparator = Selenium2Value_TableWithHeaderComparator(self.inspector)
                else:
                    obtained = self.get_table(ui_context)
                    comparator = Selenium2Value_TableComparator(inspector=self.inspector)
                res = comparator.equals(obtained, expected, raise_exception=raise_exception)
                
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Compared tables -> {res}")
                return res
            
        redo = CheckRedo()
        return redo.execute()
    
    def _get_web_element(self, element_holder):
        if element_holder.is_web_element:
            return element_holder.element
        else:
            raise TechnicalException(f"Holded element is of type '{Typing.get_object_class_fullname(element_holder.element)}' rather than WebElement")
    
    
    
    


# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import typing
import multiprocessing.context
import multiprocessing.queues
from holado_python.standard_library.queue import IterableQueue

logger = logging.getLogger(__name__)


class IterableJoinableQueue(multiprocessing.queues.JoinableQueue, typing.Iterable):
    
    __sentinel = IterableQueue.Sentinel()
    
    def __init__(self, maxsize=0, block_on_get=True, block_on_get_timeout=None, block_on_put=True, block_on_put_timeout=None):
        super().__init__(maxsize if maxsize else 0, ctx=multiprocessing.context._default_context.get_context())
        self.__block_on_get = block_on_get if block_on_get is not None else True
        self.__block_on_get_timeout = block_on_get_timeout
        self.__block_on_put = block_on_put if block_on_put is not None else True
        self.__block_on_put_timeout = block_on_put_timeout

    def __iter__(self):
        return iter(self.get, self.__sentinel)

    def __len__(self):
        return self.qsize()

    def close(self):
        self.put(self.__sentinel)
            
    def get(self):
        res = super().get(block=self.__block_on_get, timeout=self.__block_on_get_timeout)
        
        # Put again the sentinel in case of multiple readers
        if self.is_sentinel(res):
            self.put(res)
            
        return res
    
    def put(self, item):
        return super().put(item, block=self.__block_on_put, timeout=self.__block_on_put_timeout)
    
    def is_sentinel(self, item):
        return item == self.__sentinel
    
    
    
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import int
import logging
from datetime import datetime
import string
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)



class UniqueValueManager(object):
    """
    @summary: Unique value manager.
    An internal counter is used to generate a unique integer at any moment.
    At first use, this counter is initialized with the number of seconds since 01/01/2020, to ensure a minimal uniqueness between sessions.
    
    This manager can generate unique integers and unique strings.
    For strings, a unique integer is encoded in a base in range [2, 62] (default: 62).
    """
    
    def __init__(self, padding_character='0'):
        self.__last_unique_int = None
        self.__padding_character = padding_character
        
        self.__digs = string.digits + string.ascii_uppercase + string.ascii_lowercase
        
    def new_integer(self, do_log=True):
        if self.__last_unique_int is None:
            # Compute timestamp as number of seconds since 01/01/2016
            self.__last_unique_int = int((DateTime.now(tz=None) - datetime(2020, 1, 1)).total_seconds()) - 1
            
        self.__last_unique_int += 1
        res = self.__last_unique_int
        
        if do_log:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"New unique integer: {res}")
        return res
    
    def new_hex(self, length=None, truncate_to_length=True, do_log=True):
        res = self.new_string(base=16, padding_length=length, padding_character="0", raise_if_padding_impossible=not truncate_to_length, do_log=False)
        res = res.upper()
        if length is not None and len(res) > length:
            if truncate_to_length:
                res = res[-length:]
            else:
                raise TechnicalException(f"Unique HEX integer ({res}) has length {len(res)} that exceeds expected length {length}")
        
        if do_log:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"New unique HEX integer: {res}")
        return res
    
    def new_string(self, base=62, padding_length=None, padding_character=None, raise_if_padding_impossible=True, do_log=True):
        if base < 2 or base > 62:
            raise TechnicalException("Base must be in range [2, 62]")
        value = self.new_integer(do_log=False)
        res = self.__encode_int_until_base62(value, base)
        
        if padding_length:
            if len(res) > padding_length and raise_if_padding_impossible:
                raise TechnicalException(f"Unique string before padding has length {len(res)}, it is not possible to apply a padding length of {padding_length}")
            if len(res) < padding_length:
                if padding_character is None and self.__padding_character is None:
                    raise TechnicalException(f"In order to manage padding, the padding character must be given as parameter to UniqueValueManager.new_string or constructor of UniqueValueManager")
                pc = padding_character if padding_character is not None else self.__padding_character
                res = res.rjust(padding_length, pc)
            
        if do_log:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"New unique string: [{res}]")
        return res
    
    def __encode(self, int_value, base):
        # Encode given integer in 36 base
        return self.__encode_int_until_base62(int_value, base)
    
    def __encode_int_until_base62(self, x, base):
        if x < 0: 
            sign = -1
        elif x == 0: 
            return self.__digs[0]
        else: 
            sign = 1
        x *= sign
        digits = []
        while x:
            x, ind_digits = divmod(x, base)
            digits.append(self.__digs[ind_digits])
        if sign < 0:
            digits.append('-')
        digits.reverse()
        return ''.join(digits)





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
import copy
from holado.common.handlers.undefined import undefined_argument

logger = logging.getLogger(__name__)



class PersistedDataManager():
    """
    Manage data persisted in a dedicated table.
    """
    def __init__(self, data_name, table_name, table_sql_create, db_name="default", do_audit=False):
        self.__data_name = data_name
        self.__table_name = table_name
        self.__table_sql_create = table_sql_create
        self.__db_name = db_name
        self.__do_audit = do_audit
        
        self.__resource_manager = None
    
    def initialize(self, resource_manager):
        self.__resource_manager = resource_manager
        
    @property
    def table_name(self):
        return self.__table_name
    
    def does_persistent_db_exist(self):
        return self.__resource_manager.has_data_table(self.__table_name, db_name=self.__db_name, is_persistent=True) \
            and self.__resource_manager.check_data_table_schema(self.__table_name, self.__table_sql_create, db_name=self.__db_name, is_persistent=True)
    
    def ensure_persistent_db_exists(self):
        if self.__resource_manager.has_data_table(self.__table_name, db_name=self.__db_name, is_persistent=True):
            if self.__resource_manager.check_data_table_schema(self.__table_name, self.__table_sql_create, db_name=self.__db_name, is_persistent=True):
                # Table already exists with the right schema
                return
            else:
                # Table already exists but with wrong schema => delete table before creating it again
                self.__resource_manager.delete_data_table(self.__table_name, db_name=self.__db_name, is_persistent=True, raise_if_not_exist=True, do_commit=True)
        
        # Create table
        # Note: method create_data_table raise an exception if it doesn't succeed to create the table
        self.__resource_manager.create_data_table(self.__table_name, self.__table_sql_create, db_name=self.__db_name, is_persistent=True, raise_if_exist=True, do_commit=True, do_audit=self.__do_audit)
    
    def has_persisted_data(self, filter_data=None, filter_compare_data=None):
        return self.__resource_manager.has_data(self.__table_name, where_data=filter_data, where_compare_data=filter_compare_data, db_name=self.__db_name, is_persistent=True)
        
    def get_persisted_datas(self, filter_data=None, filter_compare_data=None, as_generator=False):
        """
        Note: Whereas 'data' is already a plural, a 's' is added in method name to be coherent with other method names
        """
        return self.__resource_manager.get_data(self.__table_name, where_data=filter_data, where_compare_data=filter_compare_data, db_name=self.__db_name, is_persistent=True, result_as_dict_list=True, as_generator=as_generator)

    def get_persisted_data(self, filter_data=None, filter_compare_data=None):
        """
        Note: Whereas 'datum' should be the right word in method name since method returns only one datum, method is named with 'data' in its usual singular meaning for most people.
        """
        data = self.get_persisted_datas(filter_data=filter_data, filter_compare_data=filter_compare_data)
        if len(data) > 1:
            raise TechnicalException(f"Too many ({len(data)}) {self.__data_name} found for filter {filter_data}.")
        elif len(data) == 1:
            return data[0]
        else:
            return None
        
    def count_persisted_data(self, filter_data=None, filter_compare_data=None):
        return self.__resource_manager.count_data(self.__table_name, where_data=filter_data, where_compare_data=filter_compare_data, db_name=self.__db_name, is_persistent=True)

    def update_persisted_data(self, data, filter_data=None, filter_compare_data=None):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Update persisted {self.__data_name} for {filter_data} and {filter_compare_data}: {data}")
        self.__resource_manager.update_data(self.__table_name, data, where_data=filter_data, where_compare_data=filter_compare_data, db_name=self.__db_name, is_persistent=True)

    def add_persisted_data(self, data, persisted_filter_data=None):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Add persisted {self.__data_name} for {persisted_filter_data}: {data}")
        data = copy.copy(data)
        if persisted_filter_data:
            data.update(persisted_filter_data)
        self.__resource_manager.add_data(self.__table_name, data, db_name=self.__db_name, is_persistent=True)
        
    def update_or_add_persisted_data(self, persisted_filter_data, data, persisted_data=undefined_argument):
        if persisted_data is undefined_argument:
            has_persisted_data = self.has_persisted_data(persisted_filter_data)
        else:
            has_persisted_data = persisted_data is not None
        
        if has_persisted_data:
            self.update_persisted_data(data, filter_data=persisted_filter_data)
        else:
            self.add_persisted_data(data, persisted_filter_data)
        
    def delete_persisted_data(self, filter_data):
        self.__resource_manager.delete_data(self.__table_name, where_data=filter_data, db_name=self.__db_name, is_persistent=True)







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.context.session_context import SessionContext
import logging
from holado_core.common.resource.table_data_manager import TableDataManager
from holado_python.common.tools.datetime import DateTime, TIMEZONE_LOCAL,\
    DurationUnit
import os
from holado_system.system.filesystem.file import File
import re
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.holado_config import Config
from holado_xml.xml.stream_xml_file import StreamXMLFile
import json
from holado_core.common.tools.tools import Tools
import sys
from holado_python.common.enums import ArithmeticOperator
from datetime import datetime


logger = logging.getLogger(__name__)



class CampaignManager(object):
    """ Manage all campaigns
    """
    
    def __init__(self, db_name="campaigns"):
        super().__init__()
        
        self.__db_name = db_name
        self.__resource_manager = None
        
        self.__campaigns_table_manager = TableDataManager('campaign', 'campaigns', self.__get_campaigns_table_sql_create(), db_name=self.__db_name)
        self.__campaign_scenarios_table_manager = TableDataManager('campaign scenario', 'campaign_scenarios', self.__get_campaign_scenarios_table_sql_create(), db_name=self.__db_name)
    
    def initialize(self, resource_manager):
        self.__resource_manager = resource_manager
        
        self.__campaigns_table_manager.initialize(resource_manager)
        self.__campaigns_table_manager.ensure_db_exists()
        self.__campaign_scenarios_table_manager.initialize(resource_manager)
        self.__campaign_scenarios_table_manager.ensure_db_exists()
    
    def __get_db_client(self):
        return self.__resource_manager.get_db_client(self.__db_name)
    
    def __get_campaigns_table_sql_create(self):
        return """CREATE TABLE campaigns (
                id INTEGER PRIMARY KEY,
                name TEXT NOT NULL,
                report_path TEXT NOT NULL,
                updated_at TEXT
            )"""

    def __get_campaign_scenarios_table_sql_create(self):
        return """CREATE TABLE campaign_scenarios (
                id INTEGER PRIMARY KEY,
                campaign_id INTEGER NOT NULL,
                name TEXT NOT NULL,
                report_path TEXT,
                status TEXT,
                status_at TEXT,
                details TEXT
            )"""

    def update_stored_campaigns(self):
        """ Update stored reports in DB with new campaigns
        """
        # Get report paths of campaigns to import
        dt_last_camp = self.__get_last_campaign_updated_datetime()
        dt_ref = DateTime.apply_delta_on_datetime(dt_last_camp, ArithmeticOperator.Addition, 1, DurationUnit.Microsecond) if dt_last_camp else None
        report_paths = self.__get_campaigns_report_paths(since_datetime=dt_ref)
        
        # Sort reports in time order
        report_paths = sorted(report_paths, key=lambda p: os.path.getmtime(os.path.join(p, Config.campaign_manager_import_report_name)))  # @UndefinedVariable
        logger.info(f"reports to import: {report_paths}", msg_size_limit=None)
        
        # Import reports
        for report_path in report_paths:
            self.import_campaign_reports(report_path)
    
    def __get_campaigns_report_paths(self, since_datetime):
        reports_path = SessionContext.instance().path_manager.get_reports_path(name="test_runner", with_application_group=False)
        file_paths = SessionContext.instance().path_manager.find_files(reports_path, 
                                                                       subdir_relative_path=Config.campaign_manager_import_report_name,  # @UndefinedVariable
                                                                       since_datetime=since_datetime)
        return [os.path.dirname(p) for p in file_paths]
    
    def __get_last_campaign_updated_datetime(self):
        """ From stored campaigns, return the datetime of the last updated campaign
        """
        client = self.__get_db_client()
        
        query_str = f'''
            SELECT updated_at
            FROM campaigns
            WHERE updated_at is not NULL
            ORDER BY updated_at DESC 
            LIMIT 1
        '''
        res_dict_list = client.execute(query_str, result_as_dict_list=True, as_generator=False)
        
        updated_dt_str = res_dict_list[0]['updated_at'] if res_dict_list else None
        updated_dt = DateTime.str_2_datetime(updated_dt_str, tz=TIMEZONE_LOCAL) if updated_dt_str else None
        return updated_dt
        
    def __get_campaign_updated_datetime(self, camp_id):
        data = self.__campaigns_table_manager.get_data(filter_data={'id':camp_id})
        if data is not None and DateTime.is_str_datetime(data['updated_at']):
            return DateTime.str_2_datetime(data['updated_at'], tz=TIMEZONE_LOCAL)
        else:
            return None
        
    def __update_campaign_updated_datetime(self, camp_id, updated_datetime):
        updated_dt_str = updated_datetime
        if isinstance(updated_dt_str, datetime):
            updated_dt_str = DateTime.datetime_2_str(updated_dt_str)
        self.__campaigns_table_manager.update_data({'updated_at':updated_dt_str}, filter_data={'id':camp_id})
        
    # def __get_last_campaign_scenario_status_datetime(self):
    #     """ From stored campaigns, return the datetime of the last scenario with an execution status
    #     """
    #     client = self.__get_db_client()
    #
    #     query_str = f'''
    #         SELECT status_at
    #         FROM campaign_scenarios
    #         ORDER BY status_at DESC 
    #         LIMIT 1
    #     '''
    #     res_dict_list = client.execute(query_str, result_as_dict_list=True, as_generator=False)
    #
    #     status_dt_str = res_dict_list[0]['status_at'] if res_dict_list else None
    #     status_dt = DateTime.str_2_datetime(status_dt_str, tz=TIMEZONE_LOCAL) if status_dt_str else None
    #     return status_dt
        
    def import_campaign_reports(self, report_path):
        """ Import reports of a campaign
        @param report_path Path to the campaign report
        """
        logger.info(f"Import campaign report '{report_path}'")
        
        # Add campaign
        camp_name = os.path.basename(report_path)
        camp_id = self.add_campaign_if_needed(camp_name, report_path)
        
        # Import scenario status if needed
        if self.__do_campaign_needs_import(report_path, camp_id):
            updated_dt = None
            
            # Import information from 'report_detailed_scenario.xml'
            try:
                updated_dt = self.__import_campaign_report_detailed_scenario(report_path, camp_id)
            except Exception as exc:
                if not 'No such file or directory' in str(exc):
                    logger.error(f"Failed to import campaign report '{report_path}' from '{Config.campaign_manager_import_report_name}': {Tools.represent_exception(exc)}")  # @UndefinedVariable
                    sys.exit(1)
                else:
                    logger.debug(f"Failed to import campaign report '{report_path}' from '{Config.campaign_manager_import_report_name}': {str(exc)}")  # @UndefinedVariable
            else:
                logger.debug(f"Imported campaign report '{report_path}' from '{Config.campaign_manager_import_report_name}'")  # @UndefinedVariable
            
            # Try to import information from 'report_summary_scenario_all.txt'
            if updated_dt is None:
                try:
                    updated_dt = self.__import_campaign_report_summary_scenario_all(report_path, camp_id)
                except:
                    raise TechnicalException(f"Failed to import campaign report '{report_path}'")
                else:
                    logger.debug(f"Imported campaign report '{report_path}' from 'report_summary_scenario_all.txt'")
            
            # Update campaign updated datetime
            self.__update_campaign_updated_datetime(camp_id, updated_dt)
    
    def get_scenario_history(self, scenario_name=None, size=None):
        client = self.__get_db_client()
        placeholder = client._get_sql_placeholder()
        
        # Get data from DB
        where_clause = ""
        where_data = []
        if scenario_name is not None:
            where_clause = f"where name = {placeholder}"
            where_data.append(scenario_name)
        
        query_str = f'''
            SELECT *
            FROM campaign_scenarios
            {where_clause}
            ORDER BY name, status_at DESC 
        '''
        camp_scenarios_gen = client.execute(query_str, *where_data, result_as_dict_list=True, as_generator=True)
        
        # Build result
        res = []
        cur_scenario_name = None
        cur_scenario_statuses = None
        for cs in camp_scenarios_gen:
            # Manage new scenario
            if cur_scenario_name is None or cur_scenario_name != cs['name']:
                cur_scenario_statuses = []
                cur_scenario_name = cs['name']
                res.append({'name':cur_scenario_name, 'statuses':cur_scenario_statuses})
            
            # Add campaign info for this scenario execution if size limit is not reached
            if size is None or len(cur_scenario_statuses) < size:
                cur_scenario_statuses.append({'at':cs['status_at'], 'status':cs['status']})
            
        return res
        
    def add_campaign_if_needed(self, name, report_path):
        filter_data = {'report_path': report_path}
        if not self.__campaigns_table_manager.has_data(filter_data):
            self.__campaigns_table_manager.add_data(filter_data, {'name': name})
        camp = self.__campaigns_table_manager.get_data(filter_data)
        return camp['id']
    
    def update_or_add_campaign_scenario(self, campaign_id, name, *, report_path=None, status=None, status_at_str=None, details=None):
        filter_data = {'campaign_id': campaign_id, 'name': name}
        data = {}
        if report_path is not None:
            data['report_path'] = report_path
        if status is not None:
            data['status'] = status
        if status_at_str is not None:
            data['status_at'] = status_at_str
        if details is not None:
            data['details'] = details
        
        self.__campaign_scenarios_table_manager.update_or_add_data(filter_data, data)
        # camp_sce = self.__campaign_scenarios_table_manager.get_data(filter_data)
        # return camp_sce['id']
    
    def __do_campaign_needs_import(self, report_path, camp_id):
        stored_dt = self.__get_campaign_updated_datetime(camp_id)
        if stored_dt is None:
            return True
        
        file_path = os.path.join(report_path, Config.campaign_manager_import_report_name)  # @UndefinedVariable
        if not os.path.exists(file_path):
            return True
        
        file_dt = DateTime.timestamp_to_datetime(os.path.getmtime(file_path))
        return file_dt > stored_dt
    
    def __get_campaign_status_at_by_scenario(self, camp_id):
        res = {}
        
        filter_data = {'campaign_id':camp_id}
        for data in self.__campaign_scenarios_table_manager.get_datas(filter_data, as_generator=True):
            res[data['name']] = data['status_at']
        
        return res
    
    def __import_campaign_report_detailed_scenario(self, report_path, camp_id):
        filepath = os.path.join(report_path, Config.campaign_manager_import_report_name)  # @UndefinedVariable
        if not os.path.exists(filepath):
            return None
        
        status_at_by_scenario = self.__get_campaign_status_at_by_scenario(camp_id)
        
        # Note: get file modified datetime just before reading it, so that modified datetime is not outdated and elements will be missed on next import 
        file_modified_dt = DateTime.timestamp_to_datetime(os.path.getmtime(filepath))
        with StreamXMLFile(filepath) as fin:
            scenarios = fin.read_elements_as_dict_list(force=True)      # 'force' is True to be able to read XML files with "invalid" UTF-8 characters.
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Elements read in '{filepath}':\n{scenarios}")
        
        regex_period = re.compile(r"^\[(.+) - ([^\]]+)\]$")
        for scenario_index, scenario in enumerate(scenarios):
            if 'scenario' not in scenario:
                raise TechnicalException(f"Unexpected structure of {scenario_index+1}'th scenario: {scenario}")
            scenario = scenario['scenario']
            
            name = scenario['file']
            
            m = regex_period.match(scenario['scenario_period'])
            if m:
                status_at_str = m.group(2)
                # If report is done with dates in compact format, end datetime in period is not a valid datetime, use begin datetime instead
                if not DateTime.is_str_datetime(status_at_str[:10], '%Y-%m-%d'):
                    status_at_str = m.group(1)
            else:
                raise TechnicalException(f"Failed to import report of scenario '{name}' in campaign report file '{filepath}'")
            
            if name in status_at_by_scenario and status_at_str == status_at_by_scenario[name]:
                # Scenario result is already stored
                continue
            
            scenario_report_path = scenario['report']
            status = scenario['validation_status']
            details = json.dumps( {k:scenario[k] for k in scenario.keys() if k not in ['file', 'scenario_period', 'report', 'validation_status']} )
            
            self.update_or_add_campaign_scenario(camp_id, name, report_path=scenario_report_path, status=status, status_at_str=status_at_str, details=details)
        
        return file_modified_dt
        
    def __import_campaign_report_summary_scenario_all(self, report_path, camp_id):
        file_path = os.path.join(report_path, 'report_summary_scenario_all.txt')
        if not os.path.exists(file_path):
            return None
        
        # Note: get file modified datetime just before reading it, so that modified datetime is not outdated and elements will be missed on next import 
        file_modified_dt = DateTime.timestamp_to_datetime(os.path.getmtime(file_path))
        lines = File(file_path, do_open=True, mode='rt').readlines(strip_newline=True)
        
        regex_line_period = re.compile(r"^\[(.+) - ([^\]]+)\] (.+?)(?: - .+)? - (.*)$")
        regex_line_time = re.compile(r"^(.+?) - (.+?)(?: - .+)? - (.+)$")
        regex_status = re.compile(r"^(.*?)(?: \(.*\)| => .*)?$")
        
        for line in lines:
            m = regex_line_period.match(line)
            if m:
                status_at_str = m.group(2)
                # If report is done with dates in compact format, end datetime in period is not a valid datetime, use begin datetime instead
                if not DateTime.is_str_datetime(status_at_str[:10], '%Y-%m-%d'):
                    status_at_str = m.group(1)
                scenario_name = m.group(3)
                status_info = m.group(4)
            else:
                m = regex_line_time.match(line)
                if not m:
                    raise TechnicalException(f"Unexpected line format in {file_path}: [{line}]")
                status_at_str = m.group(1)
                scenario_name = m.group(2)
                status_info = m.group(3)
            
            m = regex_status.match(status_info)
            status = m.group(1)
            
            self.update_or_add_campaign_scenario(camp_id, scenario_name, status=status, status_at_str=status_at_str)
        
        return file_modified_dt





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
import weakref
from holado_ui.ui.actors.ui_actor import UIActor
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.actors.actions import FindAction, Action
from holado_python.common.tools.comparators.string_comparator import StringComparator
from holado_core.common.actors.find_actor import FindActor

logger = logging.getLogger(__name__)



class GUIActor(UIActor):
    """ Base class for GUI actor dedicated to a GUI window.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        super().__init__(module_name)
        self.__window_weakref = None
        
        # Define some behaviors
        self.__do_write_to_element_key_by_key = False        # By default, write to element all keys in the same time
        self.__do_mouse_over_before_click = False            # By default, don't mouse over before click
    
    def initialize(self, window):
        super().initialize(window.driver)
        self.__window_weakref = weakref.ref(window)
    
    @property
    def window(self):
        res = self.__window_weakref()
        if res is None:
            raise TechnicalException("Window has been cleared")
        return res
    
    @property
    def inspector(self):
        return self.window.inspector
    
    @FindActor.default_find_builder.getter  # @UndefinedVariable
    def default_find_builder(self):
        return self.inspector.default_find_builder
    
    @FindActor.find_builder.getter  # @UndefinedVariable
    def find_builder(self):
        """
        @return Find builder from window
        """
        return self.window.find_builder
    
    @property
    def do_mouse_over_before_click(self):
        return self.__do_mouse_over_before_click
    
    @do_mouse_over_before_click.setter
    def do_mouse_over_before_click(self, mouse_over):
        """
        @param mouseOver True if mouse must be moved over the element before click on
        """
        self.__do_mouse_over_before_click = mouse_over
    
    @property
    def do_write_to_element_key_by_key(self):
        return self.__do_write_to_element_key_by_key
    
    @do_write_to_element_key_by_key.setter
    def do_write_to_element_key_by_key(self, key_by_key):
        """
        @param key_by_key True if write to element must be processed key by key
        """
        self.__do_write_to_element_key_by_key = key_by_key
    
    def click_on_element(self, element, scroll_by_x=0, scroll_by_y=0):
        """
        Click on element after having scrolled view.
        It is usually used when element is hidden by another element like a notification.
        @param element
        @param scroll_by_x
        @param scroll_by_y
        """
        raise NotImplementedError

    def write_to_element(self, element, value):
        """
        Write a text to element
        @param element Element
        @param value Text
        """
        raise NotImplementedError
    
    def act_click_on(self, scroll_by_x=0, scroll_by_y=0):
        """
        Action that click on an element
        @return Action
        @param scroll_by_x
        @param scroll_by_y
        """
        return Action("click on", lambda x: self.click_on_element(x, scroll_by_x, scroll_by_y))
        
    def act_find_and_click_on(self, finder, find_type):
        """
        Action that find element with finder and click on
        @param finder Finder
        @param find_type Find type
        @return Action
        """
        return self.act_then(self.act_find(finder, find_type), self.act_click_on())
    
    def act_find_popup(self, popup_title, redo=True):
        """
        @param popup_title Popup title
        @param redo If True, redo action
        @return Action that find popup
        """
        def func(container, candidates, find_context, find_parameters):
            return self.window.find_popup(popup_title, find_parameters.raise_no_such_element)
        res = FindAction(f"find popup '{popup_title}'", self.find_builder, func)
        return self._act_with_redo(res) if redo else res
    
    def act_check_text_is_displayed(self, ui_context, expected_str, redo=True):
        """
        Verify information displayed
        @param ui_context UI context
        @param expected_str Expected text
        @return Verify result
        """
        def func(container, candidates, find_context, find_parameters):
            # If expected doesn't contain interpret, firstly try to find directly text associated to label
            if not self.text_verifier.contains_interpret(expected_str):
                parameters_without_raise = find_parameters.without_raise()
                element = self.inspector.get_finder_text(expected_str, self.inspector.inspect_builder.context(ui_context), None).find_in(container, candidates, find_context, parameters_without_raise)
                if element is not None:
                    return True
            
            # Find text element
            element = self.inspector.get_finder_text_element(self.inspector.inspect_builder.context(ui_context), None).find_in(container, candidates, find_context, find_parameters)
            if element is None:
                # Return False in case find is done without throw
                return False
            
            
            # Verify text
            text = self.internal_api.get_element_text(element)
            comparator = StringComparator()
            return comparator.equals(text, expected_str, find_parameters.raise_no_such_element)
        
        res = FindAction(f"is displayed '{expected_str}'", self.find_builder, func)
        return self._act_with_redo(res) if redo else res

    """
    Verify information displayed
    @param ui_context UI context
    @param expected_str Expected text
    @return Verify result
    """
    def act_check_text_containing_is_displayed(self, ui_context, expected_str, redo=True):
        def func(container, candidates, find_context, find_parameters):
            element = self.inspector.get_finder_text_containing(expected_str, self.inspector.inspect_builder.context(ui_context), None).find_in(container, candidates, find_context, find_parameters)
            return element is not None
        
        res = FindAction(f"is displayed '{expected_str}'", self.find_builder, func)
        return self._act_with_redo(res) if redo else res
    
    
    
    
    


import sys
import logging
import os
import argparse

logger = None



def __insert_sys_paths():
    # Uncomment next lines to insert holado sources when using a clone of holado project rather than holado package 
    # from initialize_holado import insert_holado_source_paths
    # insert_holado_source_paths()
    
    # Insert path to folder containing initialize_script
    here = os.path.abspath(os.path.dirname(__file__))
    scripts_path = os.path.abspath(os.path.join(here, ".."))
    if scripts_path not in sys.path:
        sys.path.insert(0, scripts_path)
    
    from script_tools import insert_holado_helper_script  # @UnresolvedImport
    insert_holado_helper_script()
    
    
def __main(args, cwd):
    from holado.common.context.session_context import SessionContext
    from holado_test.behave.behave import execute_steps

    here = os.path.abspath(os.path.dirname(__file__))
    step_path = os.path.join(here, "steps")
    step_paths = [step_path]

    SessionContext.instance().behave_manager.use_independant_runner(step_paths)

    execute_steps("""
        Given DATE = datetime now
        When tester logs 'Now: ${DATE}'
        """)
    
    
def _parse_arguments():
    descr = """Independant runner
    """
    
    parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-ll', '--log-level', choices=['ERROR', 'WARNING', 'INFO', 'DEBUG', 'TRACE'], dest='log_level', default='WARNING',
                   help='Log level')
    
    parser.add_argument('-lf', '--log-in-file', dest='log_in_file', default=False, action="store_true",
                   help='If specified, log in file rather than console')
    
    res = parser.parse_args()
    
    return res


if __name__ == "__main__":
    __insert_sys_paths()
    from holado_helper.script.initialize_script import initialize  # @UnresolvedImport
    
    script_path = os.path.abspath(os.path.dirname(__file__))
    
    args = _parse_arguments()
    
    # Initialize HolAdo framework
    cwd = initialize(work_dir_path=script_path, log_level=args.log_level, log_in_file=args.log_in_file)
    
    from holado_core.common.exceptions.functional_exception import FunctionalException
    
    try:
        __main(args, cwd)
    except FunctionalException as exc:
        logging.error(f"Action is interrupted on error: {exc}")
        sys.exit(1)
    except SystemExit as exc:
        raise exc
    except:
        logging.exception("An error occured")
        sys.exit(1)
    
    sys.exit(0)
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_system.system.command.command_result import CommandResult
from holado_core.common.tools.converters.converter import Converter
import re

logger = logging.getLogger(__name__)


class CurlCommandResult(CommandResult):
    """
    Result of a cURL command
    """
    __regex_error = None
    
    def __init__(self, command, output=None, error=None, previous_command_result=None):
        super().__init__(command, output, CurlCommandResult.__extract_error(error), previous_command_result)
        self.__statistics = None

        self.__statistics = self.__extract_statistics(output, error)
        if previous_command_result is not None:
            if self.__statistics is not None and previous_command_result.statistics is not None:
                self.__statistics = previous_command_result.statistics + self.__statistics
            elif self.__statistics is None:
                self.__statistics = previous_command_result.statistics
    
    @property
    def statistics(self):
        """
        @return Command statistics
        """
        return self.__statistics
    
    @classmethod
    def __get_regex_error(cls):
        if cls.__regex_error is None:
            cls.__regex_error = re.compile(r"curl:\s+\((\d+)\)\s+(.*)")
        return cls.__regex_error
    
    @classmethod
    def __extract_error(cls, error):
        res = None
        
        if error is not None:
            m = cls.__get_regex_error().search(error)
            if m:
                res = m.group()
        
        return res

    def __extract_statistics(self, output, error):
        res = error
        
        if (output is None or len(output) == 0) and error is not None:
            m = self.__get_regex_error().search(error)
            if m:
                res = error[:m.start()]
        
        return res
    
    @property
    def error_code(self):
        """
        @return Error code if an error occurred, else -1
        """
        res = -1
        
        if self.has_error:
            m = self.__get_regex_error().search(self.error)
            if m:
                res = Converter.to_integer(m.group(1))
        
        return res
    
    @property
    def error_message(self):
        """
        @return Error message if an error occurred, else None
        """
        res = None
        
        if self.has_error:
            m = self.__get_regex_error().search(self.error)
            if m:
                res = m.group(2)
        
        return res
    
    


# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_test.behave.behave import *  # @UnusedWildImport
import time
import logging
import re
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.tools.tools import Tools
from holado_core.common.handlers.redo import Redo
from holado_test.common.exceptions.undefined_step_exception import UndefinedStepException
from holado.holado_config import Config
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()


@Step(r"wait (?P<nb_sec>{Float}) seconds")
def step_impl(context, nb_sec):
    nb_sec = StepTools.evaluate_scenario_parameter(nb_sec)
    time.sleep(nb_sec)

def _prepare_acceptable_times(accepted, timeout, polling_time):
    res_a, res_t, res_p = accepted, timeout, polling_time

    if res_t is None:
        if res_a is None:
            res_t = Config.timeout_seconds
        else:
            res_t = 10 * res_a
    if res_a is None:
        res_a = res_t
    if res_p is None:
        res_p = res_a / 100
        if res_p > 1:
            res_p = 1

    return res_a, res_t, res_p

@Step(r"wait until (?P<then_step>{AnyLazy})(?P<then_after_steps> after steps)?(?: \((?:accepted time: (?P<accepted>{Float}) s)?(?: ; )?(?:timeout: (?P<timeout>{Float}) s)?(?: ; )?(?:polling: (?P<polling>{Float}) s)?\))?")
def step_impl(context, then_step, then_after_steps, accepted, timeout, polling):
    accepted = StepTools.evaluate_scenario_parameter(accepted)
    timeout = StepTools.evaluate_scenario_parameter(timeout)
    polling = StepTools.evaluate_scenario_parameter(polling)

    before_steps = BehaveStepTools.get_step_multiline_text(context, eval_params=EvaluateParameters.nothing(), raise_exception_if_none=False)
    if then_after_steps is not None and before_steps is None:
        raise FunctionalException("Missing steps to execute before Then step")
    
    accepted, timeout, polling = _prepare_acceptable_times(accepted, timeout, polling)

    class StepRedo(Redo):
        def __init__(self, then_step, before_steps):
            super().__init__("wait until then step")
            self.__then_step = then_step
            self.__before_steps = before_steps
            self.__steps_text = format_step_with_context(context, u"Then {}".format(then_step), with_keyword=False)
            
            # Replace dynamic variable names
            self.__then_step = StepTools.replace_variable_names(self.__then_step)
            self.__before_steps = StepTools.replace_variable_names(self.__before_steps)
            self.__steps_text = StepTools.replace_variable_names(self.__steps_text)

            self.stop_retry_on(UndefinedStepException)
            self.ignoring(Exception)

        def _execute_before_process(self):
            if self.__before_steps is not None:
                execute_steps(self.__before_steps)

        def _process(self):
            execute_steps(self.__steps_text)

        def _execute_after_failure(self, exc):
            if 'UNDEFINED SUB-STEP' in str(exc):
                raise UndefinedStepException(f"A step doesn't exist in:\n{self.__steps_text}") from exc
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Following steps failed:\n{Tools.indent_string(4, self.__steps_text)}\n  Error: {Typing.get_object_class_fullname(exc)}({exc})")

        def _get_waited_description(self):
            return self.__then_step

    redo = StepRedo(then_step, before_steps)
    redo.with_accepted_time(accepted).with_timeout(timeout).polling_every(polling)
    redo.ignoring(VerifyException)
    redo.execute()

@Then(r"(?P<value>{Str}) matches pattern (?P<pattern_str>{RawStr})")
def step_impl(context, value, pattern_str):
    val = StepTools.evaluate_scenario_parameter(value)
    pattern_str = StepTools.evaluate_string_parameter(pattern_str)
    
    pattern = re.compile(pattern_str)

    if not pattern.search(val):
        raise VerifyException("Match failure, value doesn't match pattern:\n    value : {}\n  pattern : {}".format(val, pattern_str))

@Then(r"(?P<value>{Str}) doesn't match pattern (?P<pattern_str>{RawStr})")
def step_impl(context, value, pattern_str):
    val = StepTools.evaluate_scenario_parameter(value)
    pattern_str = StepTools.evaluate_string_parameter(pattern_str)
    
    pattern = re.compile(pattern_str)

    if pattern.search(val):
        raise VerifyException("Match failure, value matches pattern:\n    value : {}\n  pattern : {}".format(val, pattern_str))








import logging
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class Crc(object):
    
    @staticmethod
    def crc(msg_bytes, offset, nb_bits, crc_size, polynomial, init_val, xor_val, reverse=False):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Computing CRC for:\n  {msg_bytes=}\n  {offset=}\n  {nb_bits=}\n  {crc_size=}\n  {polynomial=}\n  {init_val=}\n  {xor_val=}")
        res = init_val
        
        # For each data bit in the data, process bits from left to right (or from right to left if reverse=True)
        if reverse:
            bits_range = range(offset+nb_bits-1, offset-1, -1)
        else:
            bits_range = range(offset, offset+nb_bits)
        for i in bits_range:
            bit = (msg_bytes[i // 8] >> (7 - (i % 8))) & 0x1
            out = (res >> (crc_size - 1)) & 0x1
            res = (res << 1) & (0xFFFFFFFF >> (32 - crc_size))
            # Apply polynomial if the bit about to be shifted out of the crc is different than the processed bit
            if bit ^ out == 1:
                res ^= polynomial
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Computing CRC: {i=}, {bit=} => {res}")
        
        res ^= xor_val
        
        # Be careful, for a crc-32, we can't make <<32, this should result to <<0
        res &= 0xFFFFFFFF >> (32 - crc_size)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"CRC for:\n  {msg_bytes=}\n  {offset=}\n  {nb_bits=}\n  {crc_size=}\n  {polynomial=}\n  {init_val=}\n  {xor_val=}\n => {res}")
        return res
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import socket
from holado_python.standard_library.socket.socket import SocketClient, SocketServer, Socket
import abc
from holado_multitask.multithreading.loopfunctionthreaded import LoopFunctionThreaded
import time
from holado.common.handlers.undefined import undefined_argument

logger = logging.getLogger(__name__)


##########################################################################
## Clients
##########################################################################


class BlockingSocketClient(SocketClient):
    """
    Base class for blocking socket client.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument, do_run_with_recv=True, do_run_with_send=True):
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay, do_run_with_recv=do_run_with_recv, do_run_with_send=do_run_with_send)
    
    def start(self, *, read_bufsize=1024, read_kwargs=None, write_kwargs=None):
        """Start client event loop.
        """
        if self.is_with_ssl:
            # Set a small timeout, so that recv has a behaviour similar flag MSG_DONTWAIT
            # See _process_recv_send implementation, recv flags are needed to be forced to 0
            if self._idle_sleep_delay is not None:
                self.internal_socket.settimeout(self._idle_sleep_delay)
            else:
                self.internal_socket.settimeout(0.01)
        
        kwargs = {'read_bufsize':read_bufsize, 'read_kwargs':read_kwargs, 'write_kwargs':write_kwargs}
        # Note: delay_between_run_sec must be None since _idle_sleep_delay is used to make sleeps
        thread = LoopFunctionThreaded(self._process_recv_send, kwargs=kwargs, register_thread=True, delay_before_run_sec=None, delay_between_run_sec=None)
        self._start_thread(thread, stop_on_close=False)
    
    def _process_recv_send(self, *, read_bufsize=1024, read_kwargs=None, write_kwargs=None):
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Processing recv & send: begin")
        has_activity = False
        read_kwargs = read_kwargs if read_kwargs is not None else {}
        write_kwargs = write_kwargs if write_kwargs is not None else {}
        
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Processing recv & send: receive data")
        recv_data = None
        if self.is_run_with_recv:
            if self.is_with_ssl:
                # ssl doesn't suppôrt flags != 0
                flags = 0
            else:
                # Add flag to not wait data
                flags = read_kwargs.get('flags', 0)
                flags |= socket.MSG_DONTWAIT
            
            try:
                recv_data = self.internal_socket.recv(read_bufsize, flags)
            except (BlockingIOError, TimeoutError):
                # No data to read
                pass
            else:
                if recv_data:
                    has_activity = True
                    with self._data_lock:
                        self._data.in_bytes += recv_data
                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug(f"[{self.name}] Received [{recv_data}] (type: {type(recv_data)} ; total: {len(self._data.in_bytes)})")
        
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Processing recv & send: send data")
        sent = None
        if self.is_run_with_send:
            with self._data_lock:
                if self._data.out_bytes:
                    has_activity = True
                    sent = self.internal_socket.send(self._data.out_bytes)
                    if sent > 0:
                        self._data.out_bytes = self._data.out_bytes[sent:]
                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug(f"[{self.name}] Sent {sent} data (remaining to send: {len(self._data.out_bytes)})")
        
        # Wait before next loop if no data was exchanged
        if not has_activity and self._idle_sleep_delay is not None:
            if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                logger.trace(f"[{self.name}] Processing recv & send: waiting {self._idle_sleep_delay} s")
            time.sleep(self._idle_sleep_delay)
            
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Processing recv & send: end")
    

class TCPBlockingSocketClient(BlockingSocketClient):
    """
    TCP socket client.
    """
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument, do_run_with_recv=True, do_run_with_send=True):
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay, do_run_with_recv=do_run_with_recv, do_run_with_send=do_run_with_send)
    
    def create_ipv4_socket(self, host, port, **kwargs):
        socket_kwargs = self._new_ssl_context_if_required(**kwargs)
        
        if self.is_with_ssl:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # sock = Socket.create_connection((host, port), do_connect=False, **socket_kwargs)
            # sock = Socket.create_connection((host, port), do_connect=True, **socket_kwargs)
            
            wrap_socket_kwargs = self._ssl_wrap_socket_kwargs
            # do_handshake_on_connect = wrap_socket_kwargs.pop('do_handshake_on_connect', True)
            do_handshake_on_connect = wrap_socket_kwargs.pop('do_handshake_on_connect', False)
            sock = self._ssl_context.wrap_socket(sock, server_hostname=host, do_handshake_on_connect=do_handshake_on_connect, **wrap_socket_kwargs)
            self._set_internal_socket(sock, is_ssl_handshake_done_on_connect=do_handshake_on_connect)
            
            sock.connect((host, port))
        else:
            sock = socket.create_connection((host, port), **socket_kwargs)
            self._set_internal_socket(sock)




##########################################################################
## Servers
##########################################################################


class BlockingSocketServer(SocketServer):
    """
    Base class for socket servers.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument):
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay)
    
    def start(self, read_bufsize=1024, *, read_kwargs=None, write_kwargs=None):
        """Start server to wait a connection, and then listen data, process data, and send result.
        Note: current implementation is simple and supports only one connection at a time.
        """
        kwargs = {'read_bufsize':read_bufsize, 'read_kwargs':read_kwargs, 'write_kwargs':write_kwargs}
        # Note: delay_between_run_sec must be None since _idle_sleep_delay is used to make sleeps
        thread = LoopFunctionThreaded(self.wait_and_process_connection, kwargs=kwargs, register_thread=True, delay_before_run_sec=None, delay_between_run_sec=None)
        self._start_thread(thread, stop_on_close=False)
    
    def wait_and_process_connection(self, read_bufsize=1024, *, read_kwargs=None, write_kwargs=None):
        """Wait a connection, and then listen data, process data, and send result.
        Note: current implementation is simple and supports only one connection at a time.
        """
        read_kwargs = read_kwargs if read_kwargs is not None else {}
        write_kwargs = write_kwargs if write_kwargs is not None else {}
        
        conn, from_addr = self.accept()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"[{self.name}] New connection: {conn} from {from_addr}")
        
        if self.is_with_ssl:
            wrap_socket_kwargs = self._ssl_wrap_socket_kwargs
            do_handshake_on_connect = wrap_socket_kwargs.pop('do_handshake_on_connect', False)
            # do_handshake_on_connect = wrap_socket_kwargs.pop('do_handshake_on_connect', True)
            ssocket = self._ssl_context.wrap_socket(conn.internal_socket, server_side=True, do_handshake_on_connect=do_handshake_on_connect)
            conn = Socket.create(name=conn.name, internal_socket=ssocket)
        
        with conn:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"[{self.name} - {from_addr}] Start read & write loop")
            while True:
                data = conn.read(read_bufsize, **read_kwargs)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{self.name} - {from_addr}] Received: {data}")
                    
                if not data:
                    break
                result = self._process_received_data(data)
                
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{self.name} - {from_addr}] Sending: {result}")
                conn.write(result, **write_kwargs)



class TCPBlockingSocketServer(BlockingSocketServer):
    """
    TCP socket server
    """
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument):
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay)
    
    def create_ipv4_socket(self, host, port, **kwargs):
        socket_kwargs = self._new_ssl_context_if_required(server_side=True, **kwargs)
        
        if self.is_with_ssl:
            # sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # sock.bind((host, port))
            # sock.listen()
            sock = socket.create_server((host, port), **socket_kwargs)
            self._set_internal_socket(sock)
        else:
            sock = socket.create_server((host, port), **socket_kwargs)
            self._set_internal_socket(sock)




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado.common.context.session_context import SessionContext
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.block.function import Function
from holado_test.scenario.step_tools import StepTools
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_multitask.multiprocessing.periodic_function_process import PeriodicFunctionProcess
from holado_multitask.multiprocessing.function_process import FunctionProcess

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_processes_manager():
    return SessionContext.instance().processes_manager

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()


@Given(r"(?P<var_name>{Variable}) = new process that calls steps(?P<is_periodic_str> periodically)?(?: \((?:delay before: (?P<delay_before_str>{Float}) s)?(?: ; )?(?:period: (?P<period_str>{Float}) s(?: ; nb times: (?P<nb_times_str>{Int}))?)?(?: ; )?(?:format with: (?P<format_with>{Variable}(?:, {Variable})*))?\))?")
def step_impl(context, var_name, is_periodic_str, delay_before_str, period_str, nb_times_str, format_with):
    """
    Create a process with given steps.
    Notes: 
        - Each created process use a copy of the variable manager. 
        - Thus the variables, defined in variable manager before process creation, can be used in the steps. 
        - But variables created or updated in the steps by the process will affect only the process itself and won't affect the original variable manager or other processes.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    delay_before_sec = StepTools.evaluate_scenario_parameter(delay_before_str)
    period_sec = StepTools.evaluate_scenario_parameter(period_str)
    nb_times = StepTools.evaluate_scenario_parameter(nb_times_str)
    format_with = StepTools.evaluate_variable_name(format_with)
    steps = BehaveStepTools.get_step_multiline_text(context, eval_params=EvaluateParameters.nothing())
    
    if is_periodic_str is not None and period_sec is None:
        raise FunctionalException(f"Period must be defined for periodic steps")
    if format_with is not None:
        steps = StepTools.format_steps_with(steps, format_with.split(', '))
    
    if period_sec is not None:
        proc_name = f"PROC_CALL_STEPS_PERIODICALLY[{var_name};{period_sec}]"
        func = PeriodicFunctionProcess(period_sec, execute_steps, args=[steps], name=proc_name, delay_before_run_sec=delay_before_sec, nb_times=nb_times)
        res = func.unique_name
    else:
        proc_name = f"PROC_CALL_STEPS[{var_name}]"
        func = FunctionProcess(execute_steps, args=[steps], name=proc_name, delay_before_run_sec=delay_before_sec)
        res = func.unique_name
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = call steps in a process(?: \((?:delay before: (?P<delay_before_str>{Float}) s)?(?: ; )?(?:format with: (?P<format_with>{Variable}(?:, {Variable})*))?\))?")
def step_impl(context, var_name, delay_before_str, format_with):
    """
    Create and start a process with given steps.
    Notes: 
        - Each created process use a copy of the variable manager. 
        - Thus the variables, defined in variable manager before process creation, can be used in the steps. 
        - But variables created or updated in the steps by the process will affect only the process itself and won't affect the original variable manager or other processes.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    delay_before_sec = StepTools.evaluate_scenario_parameter(delay_before_str)
    format_with = StepTools.evaluate_variable_name(format_with)
    steps = BehaveStepTools.get_step_multiline_text(context, eval_params=EvaluateParameters.nothing())
    
    step_params_list = []
    if delay_before_sec is not None:
        step_params_list.append(f"delay before: {delay_before_sec} s")
    if format_with is not None:
        step_params_list.append(f"format with: {format_with} s")
    step_params = " (" + " ; ".join(step_params_list) + ")" if len(step_params_list) > 0 else ""
    
    execute_steps(u"""
        Given {var_name} = new process that calls steps{step_params}
            \"""{steps}\"""
        Given start process {var_name}
        """.format(var_name=var_name, step_params=step_params, steps=BehaveStepTools.format_steps_as_parameter(steps, 12)))

@Given(r"for process (?P<var_name>{Variable}), call steps for interrupt(?: \(format with: (?P<format_with>{Variable}(?:, {Variable})*)\))?")
def step_impl(context, var_name, format_with):
    proc_id = StepTools.evaluate_variable_value(var_name)
    format_with = StepTools.evaluate_variable_name(format_with)
    steps = BehaveStepTools.get_step_multiline_text(context, eval_params=EvaluateParameters.nothing())
    
    if format_with is not None:
        steps = StepTools.format_steps_with(steps, format_with.split(', '))
        
    interrupt_func = Function(execute_steps, steps)
    
    func = __get_processes_manager().get_process(proc_id)
    if hasattr(func, "interrupt_function"):
        func.interrupt_function = interrupt_func
    else:
        raise FunctionalException("Process '{}' is not interruptable by a function".format(var_name))

@Step(r'start process (?P<var_name>{Variable})')
def step_impl(context, var_name):
    proc_id = StepTools.evaluate_variable_value(var_name)
    
    proc = __get_processes_manager().get_process(proc_id)
    proc.start()

@Step(r'join process (?P<var_name>{Variable})')
def step_impl(context, var_name):
    proc_id = StepTools.evaluate_variable_value(var_name)
    
    proc = __get_processes_manager().get_process(proc_id)
    proc.join()

@Step(r'stop process (?P<var_name>{Variable})')
def step_impl(context, var_name):
    proc_id = StepTools.evaluate_variable_value(var_name)
    
    proc = __get_processes_manager().get_process(proc_id)
    proc.interrupt()







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


#################################################
#
# Patches are done to follow ITU recommendation:
# https://www.e-navigation.nl/sites/default/files/R-REC-M.1371-5-201402-I!!PDF-E_1.pdf
#
#################################################


import logging

logger = logging.getLogger(__name__)


# Since Django 4.1, requires_system_checks cannot be a boolean anymore.
# Framework djangogrpcframework is not maintained anymore.
# Patch it by removing reference to requires_system_checks
#
# USAGE:
#    1. If not using holado in django project, copy this file in django project
#    2. In manage.py, import this file

try:
    import django_grpc_framework  # @UnusedImport
    with_django_grpc_framework = True
except:
    with_django_grpc_framework = True

if with_django_grpc_framework:
    from django_grpc_framework.management.commands.grpcrunserver import Command  # @UnusedImport
    del Command.requires_system_checks






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.handlers.object import DeleteableObject
from holado_system.system.filesystem.file import File
from holado_core.common.tools.tools import Tools
from holado.common.context.session_context import SessionContext
from holado_xml.xml.xml_2_dict import XML2Dict


def __get_path_manager():
    return SessionContext.instance().path_manager


class XMLFile(DeleteableObject):
    """
    Manage a XML file
    """

    def __init__(self, path, auto_flush=True, do_open=False, **open_kwargs):
        super().__init__(f"XML file '{path}'")

        self.__file = File(path, auto_flush=auto_flush, do_open=do_open, **open_kwargs)

    def __enter__(self):
        self.__file.__enter__()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__file.__exit__(exc_type, exc_val, exc_tb)
    
    @property
    def path(self):
        return self.__file.path

    @property
    def internal_file(self):
        return self.__file
    
    @property
    def is_open(self):
        return self.__file.is_open

    def open(self, **kwargs):
        self.__file.open(**kwargs)

    def close(self):
        self.__file.close()

    def read_as_dict(self, **parse_kwargs):
        with self.__file as fin:
            content = fin.read()
        return XML2Dict.parse(content, **parse_kwargs)

    def write_dict(self, data, pretty=True, indent=Tools.indent_string(4, ''), **unparse_kwargs):
        content = XML2Dict.unparse(data, pretty=pretty, indent=indent, **unparse_kwargs)
        with self.__file as fout:
            fout.write(content)



    @classmethod
    def read_xml_file_content_as_dict(cls, path, open_kwargs=None, parse_kwargs=None):
        open_kwargs = open_kwargs if open_kwargs is not None else {}
        parse_kwargs = parse_kwargs if parse_kwargs is not None else {}
        
        file = XMLFile(path, **open_kwargs)
        return file.read_as_dict(**parse_kwargs)

    @classmethod
    def write_xml_file_with_dict(cls, path, data, open_kwargs=None, unparse_kwargs=None):
        open_kwargs = open_kwargs if open_kwargs is not None else {}
        unparse_kwargs = unparse_kwargs if unparse_kwargs is not None else {}
        
        __get_path_manager().makedirs(path)
        
        file = XMLFile(path, **open_kwargs)
        file.write_from_dict(data, **unparse_kwargs)
        return file




